#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, Verdict, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements

@BaseEstimator.register
class SignatureFunctionSecurityEstimator(BaseEstimator):
	_ALG_NAME = "sig_fnc"

	def analyze(self, sig_fnc):
		judgements = SecurityJudgements()
		if sig_fnc.value.name == "rsa-ssa-pss":
			judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_UncommonPadding, "Not widely used padding scheme for RSA.", compatibility = Compatibility.LIMITED_SUPPORT)
		elif sig_fnc.value.name == "eddsa":
			judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_UncommonCryptosystem, "Not widely used cryptosystem.", verdict = Verdict.BEST_IN_CLASS, compatibility = Compatibility.LIMITED_SUPPORT)
		else:
			judgements += SecurityJudgement(JudgementCode.X509Cert_Signature_Function_Common, "Commonly used signature function.", commonness = Commonness.COMMON)

		return {
			"name":			sig_fnc.name,
			"pretty":		sig_fnc.value.pretty_name,
			"security":		judgements,
		}



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import pyasn1.codec.der.encoder
from pyasn1_modules import rfc2437
from x509sak.NumberTheory import NumberTheory
from x509sak.PEMDERObject import PEMDERObject
from x509sak.PublicKey import PublicKey
from x509sak.Exceptions import KeyCorruptException
from x509sak.AlgorithmDB import Cryptosystems

class RSAPrivateKey(PEMDERObject):
	"""Class that allows generate deliberately broken or oddball RSA
	keypairs."""
	_PEM_MARKER = "RSA PRIVATE KEY"
	_ASN1_MODEL = rfc2437.RSAPrivateKey

	@classmethod
	def create(cls, p, q, e = 0x10001, swap_e_d = False, valid_only = True, carmichael_totient = False):
		n = p * q
		if not carmichael_totient:
			totient = (p - 1) * (q - 1)
		else:
			totient = NumberTheory.lcm(p - 1, q - 1)
		gcd = NumberTheory.gcd(e, totient)
		if (gcd != 1) and valid_only:
			raise KeyCorruptException("e = 0x%x isnt't relative prime to totient, gcd = 0x%x. Either accept broken keys or fix e." % (e, gcd))
		d = NumberTheory.modinv(e, totient)
		if swap_e_d:
			(e, d) = (d, e)

		dmp1 = d % (p - 1)
		dmq1 = d % (q - 1)
		iqmp = NumberTheory.modinv(q, p)
		asn1 = cls._ASN1_MODEL()
		asn1["version"] = 0
		asn1["modulus"] = n
		asn1["publicExponent"] = e
		asn1["privateExponent"] = d
		asn1["prime1"] = p
		asn1["prime2"] = q
		asn1["exponent1"] = dmp1
		asn1["exponent2"] = dmq1
		asn1["coefficient"] = iqmp
		der = pyasn1.codec.der.encoder.encode(asn1)
		return cls(der)

	@property
	def cryptosystem(self):
		return Cryptosystems["RSA"]

	@property
	def pubkey(self):
		return PublicKey.create(cryptosystem = self.cryptosystem, parameters = { "n": self.n, "e": self.e })

	@property
	def p(self):
		return int(self.asn1["prime1"])

	@property
	def q(self):
		return int(self.asn1["prime2"])

	@property
	def n(self):
		return int(self.asn1["modulus"])

	@property
	def e(self):
		return int(self.asn1["publicExponent"])

	@property
	def d(self):
		return int(self.asn1["privateExponent"])

	@property
	def dmp1(self):
		return int(self.asn1["exponent1"])

	@property
	def dmq1(self):
		return int(self.asn1["exponent2"])

	@property
	def iqmp(self):
		return int(self.asn1["coefficient"])

	@property
	def phi_n(self):
		"""Returns the Euler Totient phi(n)."""
		return (self.p - 1) * (self.q - 1)

	@property
	def lambda_n(self):
		"""Returns the Carmichael Totient lambda(n)."""
		return NumberTheory.lcm(self.p - 1, self.q - 1)

	def check_integrity(self, msg = 12345678987654321):
		# Assert gcd(e, phi(n)) == 1
		phi_n = (self.p - 1) * (self.q - 1)
		gcd = NumberTheory.gcd(phi_n, self.e)
		if gcd != 1:
			raise KeyCorruptException("Expected gcd(phi(n), e) to be 1, but was %d." % (gcd))

		# Truncate msg if too large for exponent
		msg = msg % self.n

		# Calculate normale signature and verify
		sig = pow(msg, self.d, self.n)
		verify = pow(sig, self.e, self.n)
		if verify != msg:
			raise KeyCorruptException("Expected verify value %d, but got %d." % (msg, verify))

		# Test that RSA-CRT constants work
		m1 = pow(msg, self.dmp1, self.p)
		m2 = pow(msg, self.dmq1, self.q)
		sig_crt = (((self.iqmp * (m1 - m2)) % self.p) * self.q) + m2
		if sig != sig_crt:
			raise KeyCorruptException("Expected same signature for naive signature as RSA-CRT signature, but former was 0x%x and latter 0x%x." % (sig, sig_crt))

	def __str__(self):
		return "RSAPrivateKey<%d bits>" % (self.n.bit_length())



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest, ResourceFileLoader

class CmdLineTestsFindCRT(BaseTest):
	def test_find_all(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer.com.pem", "certs/ok/johannes-bauer-root.pem", "certs/ok/pubkey_sig_ed25519.pem", "certs/ok/ecc_secp256r1.pem" ]) as srcdir:
			stdout = self._run_x509sak([ "find", srcdir ]).stdout
			self.assertOcurrences(stdout, b"BEGIN CERTIFICATE", 4)

	def test_find_specific(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer.com.pem", "certs/ok/johannes-bauer-root.pem", "certs/ok/pubkey_sig_ed25519.pem", "certs/ok/ecc_secp256r1.pem" ]) as srcdir:
			stdout = self._run_x509sak([ "find", "-h", "853e", srcdir ]).stdout
			self.assertIn(b"853ecf1f70c5f7db4d3883f3f217a9a06ec29445839e9c6acbcf69a399ecaea9", stdout)
			self.assertOcurrences(stdout, b"BEGIN CERTIFICATE", 1)

	def test_find_none(self):
		with ResourceFileLoader([ "certs/ok/johannes-bauer.com.pem", "certs/ok/johannes-bauer-root.pem", "certs/ok/pubkey_sig_ed25519.pem", "certs/ok/ecc_secp256r1.pem" ]) as srcdir:
			stdout = self._run_x509sak([ "find", "-h", "abcdef112233", srcdir ]).stdout
			self.assertEqual(b"", stdout)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.OID import OIDDB
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, AnalysisOptions, Verdict, Commonness
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements
from x509sak.Tools import ValidationTools
from x509sak.FlagChecker import FlagChecker

@BaseEstimator.register
class PurposeEstimator(BaseEstimator):
	_ALG_NAME = "purpose"

	def _judge_name(self, certificate, name):
		judgements = SecurityJudgements()
		rdns = certificate.subject.get_all(OIDDB.RDNTypes.inverse("CN"))
		have_valid_cn = False
		if len(rdns) > 0:
			found_rdn = None
			for rdn in rdns:
				value = rdn.get_value(OIDDB.RDNTypes.inverse("CN"))
				if ValidationTools.validate_domainname_template_match(value.printable_value, name):
					found_rdn = rdn
					break
			if found_rdn is not None:
				if found_rdn.component_cnt == 1:
					judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_CN_Match, "Common name (CN) matches '%s'." % (name), commonness = Commonness.COMMON)
				else:
					judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_CN_MatchMultivalueRDN, "Common name (CN) matches '%s', but is part of a multi-valued RDN: %s" % (name, found_rdn.pretty_str), commonness = Commonness.HIGHLY_UNUSUAL)
			else:
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_CN_Mismatch, "No common name (CN) matches '%s'." % (name), commonness = Commonness.UNUSUAL)

		have_valid_san = False
		extension = certificate.extensions.get_first(OIDDB.X509Extensions.inverse("SubjectAlternativeName"))
		if extension is not None:
			for san_name in extension.get_all("dNSName"):
				if ValidationTools.validate_domainname_template_match(san_name.str_value, name):
					have_valid_san = True
					judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_SAN_Match, "Subject Alternative Name matches '%s'." % (name), commonness = Commonness.COMMON)
					break
			else:
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_SAN_Mismatch, "No Subject Alternative Name X.509 extension matches '%s'." % (name), commonness = Commonness.UNUSUAL)

		if (not have_valid_cn) and (not have_valid_san):
			judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_NameVerificationFailed, "Found neither valid common name (CN) nor valid subject alternative name (SAN).", commonness = Commonness.HIGHLY_UNUSUAL, verdict = Verdict.NO_SECURITY)

		return judgements

	def _judge_purpose(self, certificate, purpose):
		judgements = SecurityJudgements()
		ku_ext = certificate.extensions.get_first(OIDDB.X509Extensions.inverse("KeyUsage"))
		eku_ext = certificate.extensions.get_first(OIDDB.X509Extensions.inverse("ExtendedKeyUsage"))
		ns_ext = certificate.extensions.get_first(OIDDB.X509Extensions.inverse("NetscapeCertificateType"))

		if certificate.is_ca_certificate:
			if purpose == AnalysisOptions.CertificatePurpose.TLSServerCertificate:
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_IsCACert, "Certificate is a valid CA certificate even though it's supposed to be a TLS server.", commonness = Commonness.HIGHLY_UNUSUAL, verdict = Verdict.NO_SECURITY)
			elif purpose == AnalysisOptions.CertificatePurpose.TLSClientCertificate:
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ClientCert_IsCACert, "Certificate is a valid CA certificate even though it's supposed to be a TLS client.", commonness = Commonness.HIGHLY_UNUSUAL, verdict = Verdict.NO_SECURITY)

		if ku_ext is not None:
			# ('digitalSignature', 0),	The digitalSignature bit is asserted when the subject public key is used for verifying digital signatures, other than signatures on certificates (bit 5) and CRLs (bit 6), such as those used in an entity authentication service, a data origin authentication service, and/or an integrity service.
			# ('nonRepudiation', 1),	The nonRepudiation bit is asserted when the subject public key is used to verify digital signatures, other than signatures on certificates (bit 5) and CRLs (bit 6), used to provide a non-repudiation service that protects against the signing entity falsely denying some action.
			# ('keyEncipherment', 2),	The keyEncipherment bit is asserted when the subject public key is used for enciphering private or secret keys, i.e., for key transport.
			# ('dataEncipherment', 3),	The dataEncipherment bit is asserted when the subject public key is used for directly enciphering raw user data without the use of an intermediate symmetric cipher.  Note that the use of this bit is extremely uncommon
			# ('keyAgreement', 4),		The keyAgreement bit is asserted when the subject public key is used for key agreement.
			# ('keyCertSign', 5),
			# ('cRLSign', 6),
			# ('encipherOnly', 7),		The meaning of the encipherOnly bit is undefined in the absence of the keyAgreement bit. When the encipherOnly bit is asserted and the keyAgreement bit is also set, the subject public key may be used only for enciphering data while performing key agreement.
			# ('decipherOnly', 8)		The meaning of the decipherOnly bit is undefined in the absence of the keyAgreement bit. When the decipherOnly bit is asserted and the keyAgreement bit is also set, the subject public key may be used only for deciphering data while performing key agreement.

			flag_checker = FlagChecker()
			if purpose == AnalysisOptions.CertificatePurpose.CACertificate:
				flag_checker.may_not_have("nonRepudiation", "keyEncipherment", "dataEncipherment", "keyAgreement", "encipherOnly", "decipherOnly")
				flag_checker.must_have("keyCertSign")
				flag_checker.may_have("digitalSignature", "cRLSign")
			elif purpose in [ AnalysisOptions.CertificatePurpose.TLSClientCertificate, AnalysisOptions.CertificatePurpose.TLSServerCertificate ]:
				flag_checker.may_not_have("nonRepudiation", "dataEncipherment", "keyCertSign", "cRLSign", "encipherOnly", "decipherOnly")
				flag_checker.complex_check([ "digitalSignature", "keyEncipherment", "keyAgreement" ], min_count = 1)
			else:
				raise NotImplementedError(purpose)

			jcode = {
				"missing": {
					AnalysisOptions.CertificatePurpose.CACertificate: JudgementCode.CertUsage_Purpose_CACert_KU_MissingBits,
					AnalysisOptions.CertificatePurpose.TLSClientCertificate: JudgementCode.CertUsage_Purpose_ClientCert_KU_MissingBits,
					AnalysisOptions.CertificatePurpose.TLSServerCertificate: JudgementCode.CertUsage_Purpose_ServerCert_KU_MissingBits,
				}[purpose],
				"unusual": {
					AnalysisOptions.CertificatePurpose.CACertificate: JudgementCode.CertUsage_Purpose_CACert_KU_UnusualBits,
					AnalysisOptions.CertificatePurpose.TLSClientCertificate: JudgementCode.CertUsage_Purpose_ClientCert_KU_UnusualBits,
					AnalysisOptions.CertificatePurpose.TLSServerCertificate: JudgementCode.CertUsage_Purpose_ServerCert_KU_UnusualBits,
				}[purpose],
				"excess": {
					AnalysisOptions.CertificatePurpose.CACertificate: JudgementCode.CertUsage_Purpose_CACert_KU_ExcessBits,
					AnalysisOptions.CertificatePurpose.TLSClientCertificate: JudgementCode.CertUsage_Purpose_ClientCert_KU_ExcessBits,
					AnalysisOptions.CertificatePurpose.TLSServerCertificate: JudgementCode.CertUsage_Purpose_ServerCert_KU_ExcessBits,
				}[purpose],
			}

			for violation in flag_checker.check(ku_ext.all_flags):
				if violation.check_type == "missing":
					judgements += SecurityJudgement(jcode["missing"], "Certificate with purpose %s should have at least key usage flags %s, but %s is missing." % (purpose.name, ", ".join(sorted(violation.reference)), ", ".join(sorted(violation.flags))), commonness = Commonness.HIGHLY_UNUSUAL)
				elif violation.check_type == "excess":
					judgements += SecurityJudgement(jcode["excess"], "Certificate with purpose %s has excessive key usage bits set: %s" % (purpose.name, ", ".join(sorted(violation.flags))), commonness = Commonness.HIGHLY_UNUSUAL)
				elif violation.check_type == "unusual":
					judgements += SecurityJudgement(jcode["unusual"], "For certificate with purpose %s it is uncommon to have KeyUsage %s." % (purpose.name, ", ".join(sorted(violation.flags))), commonness = Commonness.UNUSUAL)
				elif violation.check_type == "complex_too_few":
					if len(violation.flags) == 0:
						judgements += SecurityJudgement(jcode["missing"], "Certificate with purpose %s should have at least %d flag(s) out of %s, but none were present." % (purpose.name, violation.reference_count, ", ".join(sorted(violation.reference))), commonness = Commonness.HIGHLY_UNUSUAL)
					else:
						judgements += SecurityJudgement(jcode["missing"], "Certificate with purpose %s should have at least %d flag(s) out of %s, but only %s were present." % (purpose.name, violation.reference_count, ", ".join(sorted(violation.reference)), ", ".join(sorted(violation.flags))), commonness = Commonness.HIGHLY_UNUSUAL)
				else:
					raise Exception(NotImplemented)

		if eku_ext is not None:
			if (purpose == AnalysisOptions.CertificatePurpose.TLSClientCertificate) and (not eku_ext.client_auth):
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ClientCert_EKUMismatch, "Certificate is supposed to be a client certificate and has an Extended Key Usage extension, but no clientAuth flag set within that extension.", commonness = Commonness.UNUSUAL)

			if (purpose == AnalysisOptions.CertificatePurpose.TLSServerCertificate) and (not eku_ext.server_auth):
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_EKUMismatch, "Certificate is supposed to be a server certificate and has an Extended Key Usage extension, but no serverAuth flag set within that extension.", commonness = Commonness.UNUSUAL)

		if ns_ext is not None:
			if (purpose == AnalysisOptions.CertificatePurpose.TLSClientCertificate) and (not ns_ext.ssl_client):
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ClientCert_NSCT_NoSSLClient, "Certificate is supposed to be a client certificate and has an Netscape Certificate Type extension, but no sslClient flag set within that extension.", commonness = Commonness.UNUSUAL)

			if (purpose == AnalysisOptions.CertificatePurpose.TLSServerCertificate) and (not ns_ext.ssl_server):
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_ServerCert_NSCT_NoSSLServer, "Certificate is supposed to be a server certificate and has an Netscape Certificate Type extension, but no sslServer flag set within that extension.", commonness = Commonness.UNUSUAL)

			if (purpose == AnalysisOptions.CertificatePurpose.CACertificate):
				if not any(flag in ns_ext.flags for flag in [ "sslCA", "emailCA", "objCA" ]):
					# No CA bit is set
					judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_CACert_NSCT_NoCA, "Certificate is supposed to be a CA certificate and has an Netscape Certificate Type extension, but neither sslCA/emailCA/objCA flag set within that extension.", commonness = Commonness.UNUSUAL)
				else:
					# At least it's some type of CA. But is it a SSL CA?
					if "sslCA" not in ns_ext.flags:
						judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_CACert_NSCT_NoSSLCA, "Certificate is supposed to be a CA certificate and has an Netscape Certificate Type extension, but it is not marked as an SSL CA. It can only be used for S/MIME or object signing.", commonness = Commonness.UNUSUAL)

		if purpose == AnalysisOptions.CertificatePurpose.CACertificate:
			if not certificate.is_ca_certificate:
				judgements += SecurityJudgement(JudgementCode.CertUsage_Purpose_CACert_NoCACert, "Certificate is not a valid CA certificate even though it's supposed to be.", commonness = Commonness.HIGHLY_UNUSUAL, verdict = Verdict.NO_SECURITY)

		return judgements

	def analyze(self, certificate):
		result = [ ]

		if self._analysis_options.fqdn is not None:
			analysis = {
				"type":			"name_match",
				"name":			self._analysis_options.fqdn,
				"security":		self._judge_name(certificate, self._analysis_options.fqdn),
			}
			result.append(analysis)

		for purpose in self._analysis_options.purposes:
			analysis = {
				"type":			"purpose_match",
				"purpose":		purpose,
				"security":		self._judge_purpose(certificate, purpose),
			}
			result.append(analysis)

		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.Intervals import IntervalConstraintException, IntervalRelation, Interval, Intervals

class IntervalTests(BaseTest):
	def test_interval_comparison(self):
		self.assertLess(Interval(0, 1), Interval(0, 2))
		self.assertLess(Interval(0, 1), Interval(1, 2))
		self.assertEqual(Interval(0, 10), Interval(0, 10))
		self.assertEqual(Interval(10, 20), Interval(10, 20))
		self.assertNotEqual(Interval(0, 1), Interval(0, 2))

	def assertIntervalRelation(self, intvl1, intvl2, relation):
		self.assertEqual(intvl1.relation_to(intvl2), relation)
		inv_relation = {
			IntervalRelation.Contained:	IntervalRelation.Container,
			IntervalRelation.Container:	IntervalRelation.Contained,
		}.get(relation, relation)
		self.assertEqual(intvl2.relation_to(intvl1), inv_relation)

	def test_interval_relation_disjunct(self):
		self.assertIntervalRelation(Interval(0, 10), Interval(100, 200), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(100, 200), Interval(0, 10), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(0, 10), Interval(10, 20), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(10, 20), Interval(0, 10), IntervalRelation.Disjunct)

	def test_interval_relation_identical(self):
		self.assertIntervalRelation(Interval(0, 10), Interval(0, 10), IntervalRelation.Identical)
		self.assertIntervalRelation(Interval(20, 100), Interval(20, 100), IntervalRelation.Identical)

	def test_interval_relation_overlapping(self):
		self.assertIntervalRelation(Interval(0, 10), Interval(1, 11), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(1, 11), Interval(0, 10), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(5, 200), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 5), IntervalRelation.Overlapping)

		self.assertIntervalRelation(Interval(0, 10), Interval(-100, -20), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 0), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 1), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 5), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 9), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 10), IntervalRelation.Contained)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 11), IntervalRelation.Contained)
		self.assertIntervalRelation(Interval(0, 10), Interval(-100, 1000), IntervalRelation.Contained)

		self.assertIntervalRelation(Interval(0, 10), Interval(20, 100), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(0, 10), Interval(10, 100), IntervalRelation.Disjunct)
		self.assertIntervalRelation(Interval(0, 10), Interval(9, 100), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(5, 100), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(1, 100), IntervalRelation.Overlapping)
		self.assertIntervalRelation(Interval(0, 10), Interval(0, 100), IntervalRelation.Contained)
		self.assertIntervalRelation(Interval(0, 10), Interval(-1, 100), IntervalRelation.Contained)
		self.assertIntervalRelation(Interval(0, 10), Interval(-1000, 100), IntervalRelation.Contained)

	def test_interval_enumeration(self):
		self.assertEqual(list(Interval(0, 5).enumerate_members()), [ 0, 1, 2, 3, 4 ])
		self.assertEqual(list(Interval(-3, 3).enumerate_members()), [ -3, -2, -1, 0, 1, 2 ])

	def test_interval_relation_by_enumeration(self):
		for start1 in range(-5, 5):
			for len1 in range(1, 4):
				intvl1 = Interval(start1, start1 + len1)
				members1 = set(intvl1.enumerate_members())
				for start2 in range(-5, 5):
					for len2 in range(1, 4):
						intvl2 = Interval(start2, start2 + len2)
						members2 = set(intvl2.enumerate_members())
						intersect = members1 & members2
						if members1 == members2:
							expect_relation = IntervalRelation.Identical
						elif len(intersect) == 0:
							expect_relation = IntervalRelation.Disjunct
						elif intersect == members1:
							expect_relation = IntervalRelation.Contained
						elif intersect == members2:
							expect_relation = IntervalRelation.Container
						else:
							expect_relation = IntervalRelation.Overlapping
						self.assertIntervalRelation(intvl1, intvl2, expect_relation)

	def test_intervals_enumerate(self):
		intervals = Intervals()
		intervals.add(Interval(3, 5))
		intervals.add(Interval(0, 4))
		self.assertEqual(list(intervals.enumerate_members()), [ 0, 1, 2, 3, 3, 4 ])

	def test_intervals_interacting(self):
		intervals = Intervals()
		intervals.add(Interval(0, 4))
		intervals.add(Interval(3, 6))

		self.assertEqual(list(intervals.find_interacting(Interval(-10, -5))), [ ])
		self.assertEqual(list(intervals.find_interacting(Interval(-10, 0))), [ ])
		self.assertEqual(list(intervals.find_interacting(Interval(-10, 1))), [ intervals[0] ])
		self.assertEqual(list(intervals.find_interacting(Interval(-10, 3))), [ intervals[0] ])
		self.assertEqual(list(intervals.find_interacting(Interval(-10, 4))), [ intervals[0], intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(-10, 5))), [ intervals[0], intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(-10, 10))), [ intervals[0], intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(0, 10))), [ intervals[0], intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(1, 10))), [ intervals[0], intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(3, 10))), [ intervals[0], intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(4, 10))), [ intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(5, 10))), [ intervals[1] ])
		self.assertEqual(list(intervals.find_interacting(Interval(6, 10))), [ ])
		self.assertEqual(list(intervals.find_interacting(Interval(2, 5))), [ intervals[0], intervals[1] ])

	def test_fully_contained(self):
		intervals = Intervals()
		intervals.add(Interval(0, 4))
		intervals.add(Interval(3, 6))

		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(1, 2)))
		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(1, 4)))
		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(0, 4)))
		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(0, 3)))
		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(3, 4)))
		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(3, 6)))
		self.assertTrue(intervals.fully_contained_in_subinterval(Interval(4, 5)))
		self.assertFalse(intervals.fully_contained_in_subinterval(Interval(0, 10)))
		self.assertFalse(intervals.fully_contained_in_subinterval(Interval(0, 5)))
		self.assertFalse(intervals.fully_contained_in_subinterval(Interval(-1, 4)))
		self.assertFalse(intervals.fully_contained_in_subinterval(Interval(3, 1000)))
		self.assertFalse(intervals.fully_contained_in_subinterval(Interval(-100, 1000)))

	def test_no_constraints(self):
		intervals = Intervals()
		intervals.add(Interval(0, 4))
		intervals.add(Interval(0, 4))
		intervals.add(Interval(-10, 10))

	def test_no_identical_constraint(self):
		intervals = Intervals(allow_identical = False)
		intervals.add(Interval(0, 4))
		with self.assertRaises(IntervalConstraintException):
			intervals.add(Interval(0, 4))
		intervals.add(Interval(-10, 10))

	def test_no_overlapping_constraint(self):
		intervals = Intervals(allow_overlapping = False)
		intervals.add(Interval(0, 4))
		with self.assertRaises(IntervalConstraintException):
			intervals.add(Interval(0, 4))
		with self.assertRaises(IntervalConstraintException):
			intervals.add(Interval(-10, 10))

	def test_interval_fully_contained_subinterval(self):
		intervals = Intervals()
		intervals.add(Interval(100, 1285))
		new = Interval(287, 407)
		self.assertTrue(intervals.fully_contained_in_subinterval(new))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
from x509sak.tests import BaseTest
from x509sak.tls.Structure import Structure, instantiate_member as IM, IncompleteUnpackingException, UnexpectedFixedValueException, DeserializationException
from x509sak.tls.DataBuffer import DataBuffer
from x509sak.Exceptions import ProgrammerErrorException, InvalidInputException

class _FooEnum(enum.IntEnum):
	Foo = 123
	Bar = 234
	Moo = 99

class StructureTests(BaseTest):
	_BASE_STRUCT = Structure((
		IM("first", "uint8"),
		IM("second", "uint16"),
		IM("third", "uint24"),
	), name = "TestStructure")

	def test_basic_packing(self):
		self.assertEqual(self._BASE_STRUCT.pack({
			"first":	0xaa,
			"second":	0xabcd,
			"third":	0x112233,
		}), bytes.fromhex("aa ab cd 11 22 33"))

	def test_packing_errors(self):
		with self.assertRaises(ProgrammerErrorException):
			# Missing member
			self._BASE_STRUCT.pack({
				"first":	0xaa,
				"third":	0x112233,
			})
		with self.assertRaises(InvalidInputException):
			# Out of bounds value
			self._BASE_STRUCT.pack({
				"first":	256,
				"second":	0xabcd,
				"third":	0x112233,
			})
		with self.assertRaises(ProgrammerErrorException):
			# Duplicate member name
			Structure((
				IM("data", "uint8"),
				IM("foo", "uint8"),
				IM("data", "uint8"),
			))

	def test_basic_unpacking(self):
		self.assertEqual(self._BASE_STRUCT.unpack(DataBuffer.fromhex("aa ab cd 11 22 33")), {
			"first":	0xaa,
			"second":	0xabcd,
			"third":	0x112233,
		})

	def test_position_advanced(self):
		db = DataBuffer.fromhex("aa ab cd 11 22 33")
		self._BASE_STRUCT.unpack(db)
		self.assertEqual(db.offset, 6)

	def test_position_unchanged(self):
		# Either unpacking works and offset is advanced or it remains the same
		# after unpacking, no halfways things.
		db = DataBuffer.fromhex("00 aa ab cd 11 22")
		db.offset = 1
		with self.assertRaises(DeserializationException):
			self._BASE_STRUCT.unpack(db)
		self.assertEqual(db.offset, 1)

	def test_opaque_packing(self):
		structure = Structure((IM("data", "opaque8"), ))
		self.assertEqual(structure.pack({ "data": b"foobar" }), b"\x06foobar")

		structure = Structure((IM("data", "opaque16"), ))
		self.assertEqual(structure.pack({ "data": b"foobar999" }), b"\x00\x09foobar999")

		structure = Structure((IM("data", "opaque24"), ))
		self.assertEqual(structure.pack({ "data": b"foobar999123" }), b"\x00\x00\x0cfoobar999123")

	def test_opaque_unpacking(self):
		structure = Structure((IM("data", "opaque8"), ))
		self.assertEqual(structure.unpack(DataBuffer(b"\x06foobar")), { "data": b"foobar" })

		structure = Structure((IM("data", "opaque16"), ))
		self.assertEqual(structure.unpack(DataBuffer(b"\x00\x09foobar123trail")), { "data": b"foobar123" })

		structure = Structure((IM("data", "opaque24"), ))
		self.assertEqual(structure.unpack(DataBuffer(b"\x00\x00\x0cfoobar123321blubb")), { "data": b"foobar123321" })

	def test_array_packing(self):
		structure = Structure((
			IM("int1", "uint8"),
			IM("data", "array[6]"),
			IM("int2", "uint8"),
		))
		self.assertEqual(structure.pack({
			"int1": 0xaa,
			"data": b"foobar",
			"int2": 0xbb,
		}), bytes.fromhex("aa") + b"foobar" + bytes.fromhex("bb"))

		with self.assertRaises(InvalidInputException):
			# No padding defined, must fail
			structure.pack({
				"int1": 0xaa,
				"data": b"foo",
				"int2": 0xbb,
			})

		# With padding byte
		structure = Structure((
			IM("int1", "uint8"),
			IM("data", "array[6, ab]"),
			IM("int2", "uint8"),
		))
		self.assertEqual(structure.pack({
			"int1": 0xaa,
			"data": b"foobar",
			"int2": 0xbb,
		}), bytes.fromhex("aa") + b"foobar" + bytes.fromhex("bb"))
		self.assertEqual(structure.pack({
			"int1": 0xaa,
			"data": b"foo",
			"int2": 0xbb,
		}), bytes.fromhex("aa") + b"foo" + bytes.fromhex("ab ab ab bb"))

	def test_array_unpacking(self):
		structure = Structure((
			IM("int1", "uint8"),
			IM("data", "array[6, ab]"),
			IM("int2", "uint8"),
		))
		self.assertEqual(structure.unpack(DataBuffer(b"\xaafoobar\xbb")), { "data": b"foobar", "int1": 0xaa, "int2": 0xbb })

	def test_integer_enum(self):
		structure = Structure((
			IM("a", "uint8", enum_class = _FooEnum),
			IM("b", "uint16", enum_class = _FooEnum),
			IM("c", "uint32", enum_class = _FooEnum),
			IM("d", "uint32", enum_class = _FooEnum, strict_enum = True),
		))

		orig_values = {
			"a":	_FooEnum.Foo,
			"b":	_FooEnum.Moo,
			"c":	_FooEnum.Bar,
			"d":	1234,
		}
		with self.assertRaises(InvalidInputException):
			data = structure.pack(orig_values)

		orig_values["d"] = _FooEnum.Moo
		data = structure.pack(orig_values)
		db = DataBuffer(data)
		decoded_values = structure.unpack(db)
		self.assertEqual(orig_values, decoded_values)

		orig_values["b"] = 1
		data = structure.pack(orig_values)
		db = DataBuffer(data)
		decoded_values = structure.unpack(db)
		self.assertEqual(orig_values, decoded_values)

	def _assert_encoding_decoding(self, structure, values, binary_encoding):
		# Test encoding
		data = structure.pack(values)
		self.assertEqual(data, binary_encoding)

		# Test decoding
		db = DataBuffer(data)
		decoded_values = structure.unpack(db)
		self.assertEqual(values, decoded_values)

	def test_nested_structure(self):
		structure = Structure((
			IM("a", "opaque24", inner = Structure((
				IM("b", "uint16"),
				IM("c", "uint32"),
			))),
		))
		values = {
			"a": {
				"b":	0x123,
				"c":	0x456789,
			},
		}
		binary_encoding = bytes.fromhex("00 00 06 01 23 00 45 67 89")
		self._assert_encoding_decoding(structure, values, binary_encoding)

	def test_nested_array(self):
		structure = Structure((
			IM("a", "opaque24", contains_array = True, inner = Structure((
				IM("b", "uint16"),
			))),
		))
		values = {
			"a": [
				{ "b": 0x123 },
				{ "b": 0x456 },
				{ "b": 0x789 },
				{ "b": 0x987 },
			],
		}

		binary_encoding = bytes.fromhex("00 00 08 01 23 04 56 07 89 09 87")
		self._assert_encoding_decoding(structure, values, binary_encoding)

	def test_complex_nested_array(self):
		structure = Structure((
			IM("all", "opaque16", inner = Structure((
				IM("ciphers", "opaque24", contains_array = True, inner = Structure((
					IM("cipherbase", "uint8"),
					IM("cipherid", "uint8"),
				))),
				IM("blah", "opaque16", contains_array = True, inner = Structure((
					IM("foo", "uint8"),
				))),
			))),
		))
		values = {
			"all": {
				"ciphers": [
					{ "cipherbase": 0xaa, "cipherid": 0xbb },
					{ "cipherbase": 0x11, "cipherid": 0x22 },
				],
				"blah": [
					{ "foo": 9 },
				],
			}
		}

		binary_encoding = bytes.fromhex("00 0a    00 00 04  aa bb 11 22    00 01 09")
		self._assert_encoding_decoding(structure, values, binary_encoding)

	def test_string(self):
		structure = Structure((
			IM("text", "opaque16", string_encoding = "utf-8"),
		))
		values = {
			"text":	"Foobär",
		}
		binary_encoding = bytes.fromhex("00 07    46 6f 6f 62 c3 a4 72")
		self._assert_encoding_decoding(structure, values, binary_encoding)

	def test_trailing_data_error(self):
		structure = Structure((
			IM("inner_data", "opaque8", inner = Structure([
				IM("value", "uint16"),
			])),
		))
		binary_encoding = bytes.fromhex("04 11 11 22 22")
		with self.assertRaises(IncompleteUnpackingException):
			structure.unpack(DataBuffer(binary_encoding))

	def test_expect_fixed_value_error_int(self):
		structure = Structure((
			IM("inner_data", "uint16", fixed_value = 0x1234),
		))
		binary_encoding = DataBuffer(bytes.fromhex("12 34"))
		structure.unpack(binary_encoding)

		binary_encoding = DataBuffer(bytes.fromhex("12 35"))
		with self.assertRaises(UnexpectedFixedValueException):
			structure.unpack(binary_encoding)

	def test_expect_fixed_value_error_fixed(self):
		structure = Structure((
			IM("inner_data", "fixed[12 34]"),
		))
		binary_encoding = DataBuffer(bytes.fromhex("12 34"))
		structure.unpack(binary_encoding)

		binary_encoding = DataBuffer(bytes.fromhex("12 35"))
		with self.assertRaises(UnexpectedFixedValueException):
			structure.unpack(binary_encoding)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.KwargsChecker import KwargsChecker
from x509sak.AlgorithmDB import Cryptosystems

class KeySpecification():
	def __init__(self, cryptosystem, parameters = None):
		assert(isinstance(cryptosystem, Cryptosystems))
		if parameters is None:
			parameters = { }
		self._cryptosystem = cryptosystem
		self._parameters = dict(parameters)

		constraints = KwargsChecker(required_arguments = set(param[0] for param in self._cryptosystem.value.spec_parameters))
		constraints.check(parameters, hint = "keyspec for cryptosystem %s" % (self._cryptosystem.name))

	@property
	def cryptosystem(self):
		return self._cryptosystem

	@classmethod
	def from_cmdline_str(cls, text):
		text = text.lower()
		if text.startswith("rsa:"):
			parameters = { "bitlen": int(text[4:]) }
			return cls(Cryptosystems.RSA, parameters = parameters)
		elif text.startswith("ecc:"):
			parameters = { "curvename": text[4:] }
			return cls(Cryptosystems.ECC_ECDSA, parameters = parameters)
		elif text.startswith("eddsa:"):
			parameters = { "curvename": text[6:] }
			return cls(Cryptosystems.ECC_EdDSA, parameters = parameters)
		else:
			raise ValueError("Cannot interpret command line string '%s'." % (text))

	def __eq__(self, other):
		return (self.cryptosystem, self._parameters) == (other.cryptosystem, other._parameters)

	def __getitem__(self, key):
		return self._parameters[key]

	def __str__(self):
		if self._cryptosystem == Cryptosystems.RSA:
			return "RSA-%d" % (self["bitlen"])
		elif self._cryptosystem == Cryptosystems.ECC_ECDSA:
			return "ECC-%s" % (self["curvename"])
		elif self._cryptosystem == Cryptosystems.ECC_EdDSA:
			return "EdDSA-%s" % (self["curvename"])
		return "%s-%s" % (self._cryptosystem.name, str(self._parameters))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.tls.DataBuffer import DataBuffer, DataBufferException

class TLSDataBufferTests(BaseTest):
	def test_simple(self):
		db = DataBuffer()
		db += bytes.fromhex("11 22 33")
		db.append(bytes.fromhex("44 55"))
		self.assertEqual(db.get(1), bytes.fromhex("11"))
		self.assertEqual(db.get(3), bytes.fromhex("22 33 44"))
		with self.assertRaises(DataBufferException):
			db.get(3)
		self.assertEqual(db.get(1), bytes.fromhex("55"))
		with self.assertRaises(DataBufferException):
			db.get(1)

	def test_inited(self):
		db = DataBuffer(b"ABC")
		self.assertEqual(db.get(3), bytes.fromhex("41 42 43"))



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import random
from x509sak.BaseAction import BaseAction
from x509sak.PrimeDB import PrimeDB
from x509sak.RSAPrivateKey import RSAPrivateKey
from x509sak.NumberTheory import NumberTheory
from x509sak.Exceptions import UnfulfilledPrerequisitesException, InvalidInputException

class ActionGenerateBrokenRSA(BaseAction):
	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)

		if (not self._args.force) and os.path.exists(self._args.outfile):
			raise UnfulfilledPrerequisitesException("File/directory %s already exists. Remove it first or use --force." % (self._args.outfile))

		if not self._args.gcd_n_phi_n:
			self._primetype = "2msb"
			self._p_bitlen = self._args.bitlen // 2
			self._q_bitlen = self._args.bitlen - self._p_bitlen
		else:
			self._primetype = "3msb"
			self._p_bitlen = self._args.bitlen // 3
			self._q_bitlen = self._args.bitlen - (2 * self._p_bitlen) - 1

		if (self._args.close_q) and (self._p_bitlen != self._q_bitlen):
			raise UnfulfilledPrerequisitesException("Generating a close-q keypair with a %d modulus does't work, because p would have to be %d bit and q %d bit. Choose an even modulus bitlength." % (self._args.bitlen, self._p_bitlen, self._q_bitlen))

		if self._args.q_stepping < 1:
			raise InvalidInputException("q-stepping value must be greater or equal to 1, was %d." % (self._args.q_stepping))

		self._log.debug("Selecting %s primes with p = %d bit and q = %d bit.", self._primetype, self._p_bitlen, self._q_bitlen)

		self._prime_db = PrimeDB(self._args.prime_db, generator_program = self._args.generator)
		p = None
		q = None
		while True:
			if p is None:
				p = self._prime_db.get(bitlen = self._p_bitlen, primetype = self._primetype)
				q_generator = self._select_q(p)
			if q is None:
				q = next(q_generator)
			if self._args.gcd_n_phi_n:
				# q = (2 * r * p) + 1
				r = q
				q = 2 * r * p + 1
				if not NumberTheory.is_probable_prime(q):
					q = None
					continue

			# Always make p the smaller factor
			if p > q:
				(p, q) = (q, p)

			n = p * q
			if self._args.public_exponent == -1:
				e = random.randint(2, n - 1)
			else:
				e = self._args.public_exponent

			if self._args.carmichael_totient:
				totient = NumberTheory.lcm(p - 1, q - 1)
			else:
				totient = (p - 1) * (q - 1)
			gcd = NumberTheory.gcd(totient, e)
			if self._args.accept_unusable_key or (gcd == 1):
				break
			else:
				# Pair (phi(n), e) wasn't acceptable.
				self._log.debug("gcd(totient, e) was %d, retrying.", gcd)
				if self._args.public_exponent != -1:
					# Public exponent e is fixed, need to choose another q.
					if p.bit_length() == q.bit_length():
						# Can re-use q as next p
						(p, q) = (q, None)
						q_generator = self._select_q(p)
					else:
						# When they differ in length, need to re-choose both values
						(p, q) = (None, None)

		rsa_keypair = RSAPrivateKey.create(p = p, q = q, e = e, swap_e_d = self._args.switch_e_d, valid_only = not self._args.accept_unusable_key, carmichael_totient = self._args.carmichael_totient)
		rsa_keypair.write_pemfile(self._args.outfile)
		if self._args.verbose >= 1:
			diff = q - p
			print("Generated %d bit RSA key:" % (rsa_keypair.n.bit_length()))
			print("p = 0x%x" % (rsa_keypair.p))
			if not self._args.gcd_n_phi_n:
				print("q = 0x%x" % (rsa_keypair.q))
			else:
				print("q = 2 * r * p + 1 = 0x%x" % (rsa_keypair.q))
				print("r = 0x%x" % (r))

			print("phi(n) = 0x%x" % (rsa_keypair.phi_n))
			print("lambda(n) = 0x%x" % (rsa_keypair.lambda_n))
			print("phi(n) / lambda(n) = gcd(p - 1, q - 1) = %d" % (rsa_keypair.phi_n // rsa_keypair.lambda_n))
			gcd_n_phin = NumberTheory.gcd(rsa_keypair.n, rsa_keypair.phi_n)
			if gcd_n_phin == rsa_keypair.p:
				print("gcd(n, phi(n)) = p")
			else:
				print("gcd(n, phi(n)) = 0x%x" % (gcd_n_phin))
			if self._args.close_q:
				print("q - p = %d (%d bit)" % (diff, diff.bit_length()))
			print("n = 0x%x" % (rsa_keypair.n))
			print("d = 0x%x" % (rsa_keypair.d))
			print("e = 0x%x" % (rsa_keypair.e))

	def _select_q(self, p):
		if not self._args.close_q:
			while True:
				yield self._prime_db.get(bitlen = self._q_bitlen, primetype = self._primetype)
		else:
			q = p
			while True:
				q += 2 * random.randint(1, self._args.q_stepping)
				if NumberTheory.is_probable_prime(q):
					yield q



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import hashlib
import pyasn1.codec.der.decoder
from pyasn1_modules import rfc2459, rfc2437, rfc3279
from x509sak.OID import OID, OIDDB
from x509sak.PEMDERObject import PEMDERObject
from x509sak.Tools import ASN1Tools
from x509sak.KeySpecification import KeySpecification
from x509sak.Exceptions import UnknownAlgorithmException
from x509sak.CurveDB import CurveDB
from x509sak.ECCMath import EllipticCurve
from x509sak.AlgorithmDB import PublicKeyAlgorithms
from x509sak.ASN1Models import ECParameters

class PublicKey(PEMDERObject):
	_PEM_MARKER = "PUBLIC KEY"
	_ASN1_MODEL = rfc2459.SubjectPublicKeyInfo
	_HANDLERS_BY_PK_ALG = { }
	_HANDLERS_BY_CRYPTOSYSTEM = { }

	@classmethod
	def register_handler(cls, handler_class):
		public_key_algs = handler_class._PK_ALG
		assert(public_key_algs is not None)
		if not isinstance(public_key_algs, tuple):
			public_key_algs = [ public_key_algs ]
		for public_key_alg in public_key_algs:
			cls._HANDLERS_BY_PK_ALG[public_key_alg] = handler_class
			cls._HANDLERS_BY_CRYPTOSYSTEM[public_key_alg.value.cryptosystem] = handler_class
		return handler_class

	def keyid(self, hashfnc = "sha1"):
		hashval = hashlib.new(hashfnc)
		inner_key = ASN1Tools.bitstring2bytes(self.asn1["subjectPublicKey"])
		hashval.update(inner_key)
		return hashval.digest()

	@property
	def pk_alg(self):
		return self._pk_alg

	@property
	def key(self):
		return self._key

	@property
	def keyspec(self):
		return self._key.keyspec

	@property
	def key_decodable(self):
		return self.key_decoding_error is None

	@property
	def key_decoding_error(self):
		return self._key_decoding_error

	def _post_decode_hook(self):
		alg_oid = OID.from_asn1(self.asn1["algorithm"]["algorithm"])
		self._pk_alg = PublicKeyAlgorithms.lookup("oid", alg_oid)
		if self._pk_alg is None:
			self._key_decoding_error = "Unable to determine public key algorithm for OID %s." % (alg_oid)
			return

		if self._pk_alg not in self._HANDLERS_BY_PK_ALG:
			self._key_decoding_error = "Unable to determine public key handler for public key algorithm %s." % (self._pk_alg.name)
			return

		handler_class = self._HANDLERS_BY_PK_ALG[self._pk_alg]
		key_data = ASN1Tools.bitstring2bytes(self.asn1["subjectPublicKey"])
		self._key = handler_class.from_subject_pubkey_info(self._pk_alg, self.asn1["algorithm"]["parameters"], key_data)
		self._key_decoding_error = None

	@classmethod
	def create(cls, cryptosystem, parameters):
		if cryptosystem not in cls._HANDLERS_BY_CRYPTOSYSTEM:
			raise UnknownAlgorithmException("Unable to create public key for cryptosystem %s." % (cryptosystem.name))

		handler_class = cls._HANDLERS_BY_CRYPTOSYSTEM[cryptosystem]
		asn1 = handler_class.create(parameters)
		encoded_pubkey = pyasn1.codec.der.encoder.encode(asn1)
		return cls(encoded_pubkey)

	def __getattr__(self, name):
		try:
			return self._key[name]
		except KeyError:
			raise AttributeError("%s public key does not have a '%s' attribute." % (self.pk_alg.value.name, name))

	def __str__(self):
		return "PublicKey<%s>" % (self.keyspec)


class _BasePublicKey():
	_PK_ALG = None

	def __init__(self, accessible_parameters, decoding_details = None):
		self._accessible_parameters = accessible_parameters if (accessible_parameters is not None) else { }
		self._decoding_details = decoding_details

	@property
	def decoding_details(self):
		return self._decoding_details

	@property
	def malformed(self):
		raise NotImplementedError(self.__class__.__name__)

	def has_param(self, name):
		return name in self._accessible_parameters

	def _param(self, name):
		return self._accessible_parameters.get(name)

	def __getitem__(self, name):
		return self._accessible_parameters[name]

	@classmethod
	def create(cls, parameters):
		raise NotImplementedError(cls.__name__)

	@classmethod
	def from_subject_pubkey_info(cls, pk_alg, params_asn1, pubkey_data):
		raise NotImplementedError(cls.__name__)


@PublicKey.register_handler
class RSAPublicKey(_BasePublicKey):
	_PK_ALG = PublicKeyAlgorithms.RSA

	@property
	def malformed(self):
		return not self.has_param("n")

	@property
	def keyspec(self):
		return KeySpecification(cryptosystem = self._PK_ALG.value.cryptosystem, parameters = { "bitlen": self["n"].bit_length() })

	@classmethod
	def create(cls, parameters):
		asn1 = rfc2459.SubjectPublicKeyInfo()
		asn1["algorithm"] = rfc2459.AlgorithmIdentifier()
		asn1["algorithm"]["algorithm"] = OIDDB.KeySpecificationAlgorithms.inverse("rsaEncryption").to_asn1()
		asn1["algorithm"]["parameters"] = pyasn1.type.univ.Any(value = pyasn1.codec.der.encoder.encode(pyasn1.type.univ.Null()))

		inner_key = rfc2437.RSAPublicKey()
		inner_key["modulus"] = pyasn1.type.univ.Integer(parameters["n"])
		inner_key["publicExponent"] = pyasn1.type.univ.Integer(parameters["e"])
		inner_key = pyasn1.codec.der.encoder.encode(inner_key)
		asn1["subjectPublicKey"] = ASN1Tools.bytes2bitstring(inner_key)
		return asn1

	@classmethod
	def from_subject_pubkey_info(cls, pk_alg, params_asn1, pubkey_data):
		pubkey = ASN1Tools.safe_decode(pubkey_data, asn1_spec = rfc2437.RSAPublicKey())

		if pubkey.asn1 is not None:
			accessible_parameters = {
				"n":		int(pubkey.asn1["modulus"]),
				"e":		int(pubkey.asn1["publicExponent"]),
				"params":	params_asn1 if params_asn1.hasValue() else None,
			}
		else:
			accessible_parameters = None
		return cls(accessible_parameters = accessible_parameters, decoding_details = pubkey)

	def __repr__(self):
		if self._param("n") is not None:
			return "%s(%d bit modulus)" % (self.__class__.__name__, self["n"].bit_length())
		else:
			return "%s(undecodable)" % (self.__class__.__name__)


@PublicKey.register_handler
class DSAPublicKey(_BasePublicKey):
	_PK_ALG = PublicKeyAlgorithms.DSA

	@property
	def malformed(self):
		return not (self.has_param("p") and self.has_param("pubkey"))

	@property
	def N(self):
		return self._param("p").bit_length() if (self._param("p") is not None) else None

	@property
	def L(self):
		return self._param("q").bit_length() if (self._param("q") is not None) else None

	@property
	def keyspec(self):
		return KeySpecification(cryptosystem = self._PK_ALG.value.cryptosystem, parameters = { "N": self.N, "L": self.L })

	@classmethod
	def create(cls, parameters):
		asn1 = rfc2459.SubjectPublicKeyInfo()
		asn1["algorithm"] = rfc2459.AlgorithmIdentifier()
		asn1["algorithm"]["algorithm"] = OIDDB.KeySpecificationAlgorithms.inverse("id-dsa").to_asn1()
		asn1["algorithm"]["parameters"] = rfc3279.Dss_Parms()
		asn1["algorithm"]["parameters"]["p"] = parameters["p"]
		asn1["algorithm"]["parameters"]["q"] = parameters["q"]
		asn1["algorithm"]["parameters"]["g"] = parameters["g"]

		inner_key = rfc3279.DSAPublicKey(parameters["pubkey"])
		inner_key = pyasn1.codec.der.encoder.encode(inner_key)
		asn1["subjectPublicKey"] = ASN1Tools.bytes2bitstring(inner_key)

		return asn1

	@classmethod
	def from_subject_pubkey_info(cls, pk_alg, params_asn1, pubkey_data):
		params = ASN1Tools.safe_decode(params_asn1, asn1_spec = rfc3279.Dss_Parms())
		pubkey = ASN1Tools.safe_decode(pubkey_data, asn1_spec = rfc3279.DSAPublicKey())

		accessible_parameters = { }
		if params.asn1 is not None:
			accessible_parameters.update({
				"p":		int(params.asn1["p"]),
				"q":		int(params.asn1["q"]),
				"g":		int(params.asn1["g"]),
			})

		if pubkey.asn1 is not None:
			accessible_parameters.update({
				"pubkey":	int(pubkey.asn1),
			})

		return cls(accessible_parameters = accessible_parameters, decoding_details = [ params, pubkey ])

	def __repr__(self):
		if self._param("p") is not None:
			return "%s(%d-%d)" % (self.__class__.__name__, self.N, self.L)
		else:
			return "%s(undecodable)" % (self.__class__.__name__)


@PublicKey.register_handler
class ECDSAPublicKey(_BasePublicKey):
	_PK_ALG = PublicKeyAlgorithms.ECC

	@property
	def malformed(self):
		return (self._param("curve") is None) and (self._param("curve_source") != "implicitCurve")

	@property
	def keyspec(self):
		if self._param("curve_source") == "namedCurve":
			return KeySpecification(cryptosystem = self._PK_ALG.value.cryptosystem, parameters = { "curvename": self["curve"].name })
		else:
			return None

	@classmethod
	def create(cls, parameters):
		asn1 = rfc2459.SubjectPublicKeyInfo()

		asn1["algorithm"]["algorithm"] = OIDDB.KeySpecificationAlgorithms.inverse("ecPublicKey").to_asn1()
		if parameters["curve"].oid is not None:
			asn1["algorithm"]["parameters"] = pyasn1.codec.der.encoder.encode(parameters["curve"].oid.to_asn1())
		else:
			# TODO not implemented
			#domain_params = SpecifiedECDomain()
			#domain_params["version"] = 1
			#asn1["algorithm"]["parameters"] = domain_params
			raise NotImplementedError("Creation of explicitly specified elliptic curve domain parameters (i.e., non-named curves) is not implemented in x509sak")

		inner_key = parameters["curve"].point(parameters["x"], parameters["y"]).encode()
		asn1["subjectPublicKey"] = ASN1Tools.bytes2bitstring(inner_key)
		return asn1

	@classmethod
	def from_subject_pubkey_info(cls, pk_alg, params_asn1, pubkey_data):
		params = ASN1Tools.safe_decode(params_asn1, asn1_spec = ECParameters())

		accessible_parameters = { }
		if params.asn1 is not None:
			accessible_parameters["curve_source"] = params.asn1.getName()
			if accessible_parameters["curve_source"] == "namedCurve":
				# Named curve
				curve_oid = OID.from_asn1(params.asn1.getComponent())
				curve = CurveDB().instantiate(oid = curve_oid)
				accessible_parameters.update({
					"curve_oid":		curve_oid,
					"curve":			curve,
				})
			elif accessible_parameters["curve_source"] == "specifiedCurve":
				# Explicit curve or implicit curve
				curve = EllipticCurve.from_asn1(params.asn1.getComponent())
				accessible_parameters.update({
					"curve":			curve,
				})
			else:
				# Implicit curve
				pass

		if accessible_parameters.get("curve") is not None:
			pk_point = curve.decode_point(pubkey_data)
			accessible_parameters.update({
				"x": pk_point.x,
				"y": pk_point.y,
			})
		else:
			pk_point = None

		return cls(accessible_parameters = accessible_parameters, decoding_details = [ params, pk_point ])


@PublicKey.register_handler
class EdDSAPublicKey(_BasePublicKey):
	_PK_ALG = (PublicKeyAlgorithms.Ed25519, PublicKeyAlgorithms.Ed448)

	@property
	def malformed(self):
		return False

	@property
	def point(self):
		return self["curve"].point(self["x"], self["y"])

	@property
	def keyspec(self):
		# TODO 25519 or 448?
		return KeySpecification(cryptosystem = self._PK_ALG[0].value.cryptosystem, parameters = { "curvename": self["curve"].name })

	@classmethod
	def create(cls, parameters):
		asn1 = rfc2459.SubjectPublicKeyInfo()
		asn1["algorithm"] = rfc2459.AlgorithmIdentifier()
		asn1["algorithm"]["algorithm"] = OIDDB.KeySpecificationAlgorithms.inverse("id-dsa").to_asn1()
		asn1["algorithm"]["algorithm"] = parameters["curve"].oid.to_asn1()

		inner_key = parameters["curve"].point(parameters["x"], parameters["y"]).encode()
		asn1["subjectPublicKey"] = ASN1Tools.bytes2bitstring(inner_key)
		return asn1

	@classmethod
	def from_subject_pubkey_info(cls, pk_alg, params_asn1, pubkey_data):
		curve = CurveDB().instantiate(oid = pk_alg.value.oid)
		pk_point = curve.decode_point(pubkey_data)

		accessible_parameters = dict(pk_alg.value.fixed_params)
		accessible_parameters.update({
			"x":			pk_point.x,
			"y":			pk_point.y,
			"curve":		curve,
			"point":		pk_point,
			"curve_source":	"namedCurve",
		})
		return cls(accessible_parameters = accessible_parameters, decoding_details = [ pk_point ])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import tempfile
from x509sak.BaseAction import BaseAction
from x509sak.OpenSSLTools import OpenSSLTools
from x509sak import X509Certificate
from x509sak.Tools import ASN1Tools
from x509sak.PublicKey import PublicKey
from x509sak.Exceptions import InvalidInputException
from x509sak.PrivateKeyStorage import PrivateKeyStorage, PrivateKeyStorageForm
from x509sak.OID import OIDDB
from x509sak.X509Extensions import X509SubjectKeyIdentifierExtension, X509AuthorityKeyIdentifierExtension

class ActionForgeCert(BaseAction):
	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)

		certs = X509Certificate.read_pemfile(self._args.crt_filename)
		if not certs[0].is_selfsigned:
			raise InvalidInputException("First certificate in chain (%s) is not self-signed." % (certs[0]))
		for (cert_id, (issuer, subject)) in enumerate(zip(certs, certs[1:]), 1):
			if not subject.signed_by(issuer):
				raise InvalidInputException("Certificate %d in file (%s) is not issuer for certificate %d (%s)." % (cert_id, issuer, cert_id + 1, subject))

		self._log.debug("Chain of %d certificates to forge.", len(certs))
		self._forge_cert(0, certs[0], 0, certs[0])
		for (cert_subject_id, (issuer, subject)) in enumerate(zip(certs, certs[1:]), 1):
			self._forge_cert(cert_subject_id - 1, issuer, cert_subject_id, subject)

	def _forge_cert(self, cert_issuer_id, issuer, cert_subject_id, subject):
		self._log.debug("Forging chain element %d -> %d: %s -> %s", cert_issuer_id, cert_subject_id, issuer, subject)
		issuer_key_filename = self._args.key_template % (cert_issuer_id)
		key_filename = self._args.key_template % (cert_subject_id)
		crt_filename = self._args.cert_template % (cert_subject_id)

		if (not os.path.isfile(key_filename)) or self._args.force:
			OpenSSLTools.create_private_key(PrivateKeyStorage(storage_form = PrivateKeyStorageForm.PEM_FILE, filename = key_filename), subject.pubkey.keyspec)

		# Read new private key and convert to public key
		with tempfile.NamedTemporaryFile(prefix = "pubkey_", suffix = ".pem") as pubkey_file:
			OpenSSLTools.private_to_public(key_filename, pubkey_file.name)
			subject_pubkey = PublicKey.read_pemfile(pubkey_file.name)[0]

		# Do the same for the issuer key to get the issuer key ID
		with tempfile.NamedTemporaryFile(prefix = "pubkey_", suffix = ".pem") as pubkey_file:
			OpenSSLTools.private_to_public(issuer_key_filename, pubkey_file.name)
			issuer_pubkey = PublicKey.read_pemfile(pubkey_file.name)[0]

		# Replace public key first
		forged_cert_asn1 = subject.asn1_clone
		forged_cert_asn1["tbsCertificate"]["subjectPublicKeyInfo"] = subject_pubkey.asn1

		# Do identifiers need to be recalculated?
		if self._args.recalculate_keyids:
			if subject.extensions.has(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier")):
				# Replace subject key identifier
				new_key_id = subject_pubkey.keyid()
				replacement_extension = X509SubjectKeyIdentifierExtension.construct(new_key_id)
				subject.extensions.filter(OIDDB.X509Extensions.inverse("SubjectKeyIdentifier"), replacement_extension)

			if subject.extensions.has(OIDDB.X509Extensions.inverse("AuthorityKeyIdentifier")):
				# Replace authority key identifier
				new_key_id = issuer_pubkey.keyid()
				replacement_extension = X509AuthorityKeyIdentifierExtension.construct(new_key_id)
				subject.extensions.filter(OIDDB.X509Extensions.inverse("AuthorityKeyIdentifier"), replacement_extension)

			forged_cert_asn1["tbsCertificate"]["extensions"] = subject.extensions.to_asn1()

		# Re-serialize certificate
		forged_cert = X509Certificate.from_asn1(forged_cert_asn1)

		# Then sign the modified certifiate
		signature = OpenSSLTools.sign_data(subject.signature_algorithm, issuer_key_filename, forged_cert.signed_payload)

		# Finally, place the signature into the certificate
		forged_cert_asn1["signatureValue"] = ASN1Tools.bytes2bitstring(signature)
		forged_cert = X509Certificate.from_asn1(forged_cert_asn1)
		forged_cert.write_pemfile(crt_filename)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
from x509sak.tls.Structure import Structure, instantiate_member as IM
from x509sak.tls.Enums import SignatureAlgorithm, HashAlgorithm

class SCTVersion(enum.IntEnum):
	v1 = 0

class SCTSignatureType(enum.IntEnum):
	certificate_timestamp = 0
	tree_hash = 1

DigitalSignature = Structure(name = "DigitalSignature", members = [
	IM("hash_algorithm",				"uint8", enum_class = HashAlgorithm),
	IM("sig_algorithm",					"uint8", enum_class = SignatureAlgorithm),
	IM("signature",						"opaque16"),
])

SignedCertificateTimestamp = Structure(name = "SignedCertificateTimestamps", members = [
	IM("sct_version",					"uint8", enum_class = SCTVersion),
	IM("log_id",						"array[32]"),
	IM("timestamp",						"uint64"),
	IM("extensions",					"opaque16"),
	DigitalSignature,
])

SignedCertificateTimestampList = Structure(name = "SignedCertificateTimestampList", members = [
	IM("payload",						"opaque16", contains_array = True, inner = Structure([
		IM("sct",							"opaque16", inner = SignedCertificateTimestamp)
	])),
])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2021 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
import urllib
from x509sak.tests import BaseTest
from x509sak.SubprocessExecutor import SubprocessExecutor
from x509sak.Tools import PathTools
from x509sak.WorkDir import WorkDir

class SoftHSMInstance():
	_SO_SEARCH_PATH = "/usr/local/lib/softhsm:/usr/lib/softhsm:/usr/local/lib:/usr/lib:/usr/lib/x86_64-linux-gnu:/usr/lib/x86_64-linux-gnu/openssl-1.0.2/engines:/usr/lib/x86_64-linux-gnu/engines-1.1"

	def __init__(self):
		self._softhsm_config_file = None
		self._softhsm_storage_dir = None
		self._module_so = None
		self._env = None

	@property
	def so_search_path(self):
		return self._SO_SEARCH_PATH

	@property
	def env(self):
		return self._env

	@property
	def module_so(self):
		return self._module_so

	def __enter__(self):
		self._softhsm_config_file = tempfile.NamedTemporaryFile(mode = "w", prefix = "softhsm_", suffix = ".conf")
		self._softhsm_storage_dir = tempfile.TemporaryDirectory(prefix = "objects_")

		try:
			self._module_so = PathTools.find(self._SO_SEARCH_PATH, "libsofthsm2.so")
			if self._module_so is None:
				raise Exception("libsofthsm2.so cannot be found anywhere in %s" % (self._SO_SEARCH_PATH))

			print("directories.tokendir = %s" % (self._softhsm_storage_dir.name), file = self._softhsm_config_file)
			print("directories.backend = file", file = self._softhsm_config_file)
			print("log.level = INFO", file = self._softhsm_config_file)
			self._softhsm_config_file.flush()
			self._env = { "SOFTHSM2_CONF": self._softhsm_config_file.name }
			SubprocessExecutor([ "softhsm2-util", "--init-token", "--slot", "0", "--label", "TestToken", "--so-pin", "3537363231383830", "--pin", "648219" ], env = self._env).run()

			return self
		except:
			print("Fatal error: SoftHSM initialization failed.")
			self.__cleanup()
			raise

	def keygen(self, key_id, key_label, key_spec):
		SubprocessExecutor([ "pkcs11-tool", "--module", self.module_so, "--login", "--pin", "648219", "--keypairgen", "--key-type", key_spec, "--id", "%x" % (key_id), "--label", key_label ], env = self._env).run()
		return "pkcs11:object=%s;type=private;pin-value=648219;token=TestToken" % (urllib.parse.quote(key_label))

	def list_objects(self):
		output = SubprocessExecutor([ "pkcs11-tool", "--module", self.module_so, "--login", "--pin", "648219", "--list-objects" ], env = self._env).run().stdout
		print(output.decode())

	def get_pubkey_der(self, key_id):
		output = SubprocessExecutor([ "pkcs11-tool", "--module", self.module_so, "--login", "--pin", "648219", "--read-object", "--type", "pubkey", "--id", "%x" % (key_id) ], env = self._env).run().stdout
		return output

	def get_ecc_pubkey_text(self, key_id):
		pubkey = self.get_pubkey_der(key_id)
		result = SubprocessExecutor([ "openssl", "ec", "-inform", "der", "-pubin", "-text", "-noout" ], stdin = pubkey).run()
		return result.stdout_text

	def __cleanup(self):
		self._softhsm_storage_dir.cleanup()
		self._softhsm_config_file.close()

	def __exit__(self, *args):
		self.__cleanup()

class HardwareTokenTests(BaseTest):
	def test_create_ca(self):
		with SoftHSMInstance() as hsm, tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			key_name = hsm.keygen(key_id = 12345, key_label = "my secure key", key_spec = "EC:secp256r1")
			pubkey = hsm.get_ecc_pubkey_text(key_id = 12345)
			pubkey = pubkey.split("\n")
			pubkey_start = pubkey[2]
			self.assertTrue(pubkey_start.startswith("    04:"))
			pubkey_start = pubkey_start[4:].encode("ascii")

			# With unknown key object name, it fails
			self._run_x509sak([ "createca", "-f", "-s", "/CN=Root CA with key in HSM", "--pkcs11-so-search", hsm.so_search_path, "--pkcs11-module", "libsofthsm2.so", "--hardware-key", key_name.replace("secure", "UNKNOWN"), "root_ca" ], env = hsm.env, success_return_codes = [ 1 ])

			# With unknown token name, it fails
			self._run_x509sak([ "createca", "-f", "-s", "/CN=Root CA with key in HSM", "--pkcs11-so-search", hsm.so_search_path, "--pkcs11-module", "libsofthsm2.so", "--hardware-key", key_name.replace("Token", "UNKNOWN"), "root_ca" ], env = hsm.env, success_return_codes = [ 1 ])

			# Create root certificate with key in SoftHSM
			self._run_x509sak([ "createca", "-f", "-s", "/CN=Root CA with key in HSM", "--pkcs11-so-search", hsm.so_search_path, "--pkcs11-module", "libsofthsm2.so", "--hardware-key", key_name, "root_ca" ], env = hsm.env)

			# Check that it's validly self-signed
			SubprocessExecutor([ "openssl", "verify", "-check_ss_sig", "-CAfile", "root_ca/CA.crt", "root_ca/CA.crt" ]).run()

			# Check that the public key on the smart card appears inside the certificate
			output = SubprocessExecutor([ "openssl", "x509", "-noout", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(pubkey_start, output)


	def test_create_certificate(self):
		with SoftHSMInstance() as hsm, tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			key_name = hsm.keygen(key_id = 1, key_label = "CA_key", key_spec = "EC:secp256r1")

			# Create root certificate with key in SoftHSM
			self._run_x509sak([ "createca", "-s", "/CN=Root CA with key in HSM", "--pkcs11-so-search", hsm.so_search_path, "--pkcs11-module", "libsofthsm2.so", "--hardware-key", key_name, "root_ca" ], env = hsm.env)

			# Then create child certificate with that CA, but have child key in software
			self._run_x509sak([ "createcsr", "-s", "/CN=Child Cert", "-t", "tls-client", "-c", "root_ca", "client.key", "client.crt" ], env = hsm.env)

			# Verify the child certificate is valid
			SubprocessExecutor([ "openssl", "verify", "-check_ss_sig", "-CAfile", "root_ca/CA.crt", "client.crt" ]).run()

	def test_sign_csr(self):
		with SoftHSMInstance() as hsm, tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			key_name = hsm.keygen(key_id = 1, key_label = "CA_key", key_spec = "EC:secp256r1")

			# Create root certificate with key in SoftHSM
			self._run_x509sak([ "createca", "-s", "/CN=Root CA with key in HSM", "--pkcs11-so-search", hsm.so_search_path, "--pkcs11-module", "libsofthsm2.so", "--hardware-key", key_name, "root_ca" ], env = hsm.env)

			# Then create child CSR
			self._run_x509sak([ "createcsr", "client.key", "client.csr" ], env = hsm.env)

			# Finally, sign the child certificate
			self._run_x509sak([ "signcsr", "-s", "/CN=Child Cert", "-t", "tls-client", "root_ca", "client.csr", "client.crt" ], env = hsm.env)

			# Verify the child certificate is valid
			SubprocessExecutor([ "openssl", "verify", "-check_ss_sig", "-CAfile", "root_ca/CA.crt", "client.crt" ]).run()

	def test_revoke_certificate(self):
		with SoftHSMInstance() as hsm, tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			key_name = hsm.keygen(key_id = 1, key_label = "CA_key", key_spec = "EC:secp256r1")

			# Create root certificate with key in SoftHSM
			self._run_x509sak([ "createca", "-s", "/CN=Root CA with key in HSM", "--pkcs11-so-search", hsm.so_search_path, "--pkcs11-module", "libsofthsm2.so", "--hardware-key", key_name, "root_ca" ], env = hsm.env)

			# Then create child certificate with that CA, but have child key in software
			self._run_x509sak([ "createcsr", "-s", "/CN=Child Cert", "-t", "tls-client", "-c", "root_ca", "client.key", "client.crt" ], env = hsm.env)

			# Verify the child certificate is valid
			SubprocessExecutor([ "openssl", "verify", "-check_ss_sig", "-CAfile", "root_ca/CA.crt", "client.crt" ]).run()

			# Revoke the certificate
			self._run_x509sak([ "revokecrt", "root_ca", "client.crt" ], env = hsm.env)

			# Emit a CRL
			self._run_x509sak([ "createcrl", "root_ca", "revoked_client.crl" ], env = hsm.env)

			# Verify the child certificate is *not* valid anymore
			SubprocessExecutor([ "openssl", "verify", "-crl_check", "-CRLfile", "revoked_client.crl", "-check_ss_sig", "-CAfile", "root_ca/CA.crt", "client.crt" ], success_return_codes = [ 2 ]).run()



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.IPAddress import IPAddress, IPAddressSubnet
from x509sak.Exceptions import InvalidIPAddressException

class IPAddressTests(BaseTest):
	def test_ipv4(self):
		ip = IPAddress(bytes.fromhex("aa bb cc dd"))
		self.assertEqual(str(ip), "170.187.204.221")
		self.assertEqual(int(ip), 0xaabbccdd)

	def test_ipv6(self):
		ip = IPAddress(bytes.fromhex("aa bb cc dd 1a 1b 1c 1d 2a 2b 2c 2d 3a 3b 3c 3d"))
		self.assertEqual(str(ip), "aabb:ccdd:1a1b:1c1d:2a2b:2c2d:3a3b:3c3d")
		self.assertEqual(int(ip), 0xaabbccdd1a1b1c1d2a2b2c2d3a3b3c3d)

	def test_ipv6_shortlead(self):
		ip = IPAddress(bytes.fromhex("01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 00 0f"))
		self.assertEqual(str(ip), "102:304:506:708:90a:b0c:d0e:f")

	def test_ipv6_shortabbr1(self):
		ip = IPAddress(bytes.fromhex("1122 3344 0000 0000 0000 0000 0000 5566"))
		self.assertEqual(str(ip), "1122:3344::5566")

	def test_ipv6_shortabbr2(self):
		ip = IPAddress(bytes.fromhex("1122 0000 3344 0000 0000 0000 0000 5566"))
		self.assertEqual(str(ip), "1122::3344:0:0:0:0:5566")

	def test_ipv6_shortabbr3(self):
		ip = IPAddress(bytes.fromhex("0000 0000 0000 0000 0000 0000 0000 5566"))
		self.assertEqual(str(ip), "::5566")

	def test_ipv6_shortabbr4(self):
		ip = IPAddress(bytes.fromhex("1122 0000 0000 0000 0000 0000 0000 5566"))
		self.assertEqual(str(ip), "1122::5566")

	def test_ipv6_shortabbr5(self):
		ip = IPAddress(bytes.fromhex("1122 0000 0000 0000 0000 0000 0000 0000"))
		self.assertEqual(str(ip), "1122::")

	def test_ipv6_shortabbr6(self):
		ip = IPAddress(bytes.fromhex("0000 0000 0000 0000 0000 0000 0000 0000"))
		self.assertEqual(str(ip), "::")

	def test_malformed(self):
		with self.assertRaises(InvalidIPAddressException):
			IPAddress(bytes.fromhex("aa bb cc"))

	def test_subnet(self):
		subnet = IPAddressSubnet.from_bytes(bytes.fromhex("aa bb cc 00 ff ff ff 00"))
		self.assertEqual(str(subnet), "170.187.204.0/24")

	def test_subnet_overlap(self):
		subnet = IPAddressSubnet.from_bytes(bytes.fromhex("aa bb cc 00 ff ff aa 00"))
		self.assertEqual(str(subnet), "170.187.204.0/255.255.170.0")

	def test_subnet_noncidr(self):
		subnet = IPAddressSubnet.from_bytes(bytes.fromhex("aa bb 00 cc ff ff 00 ff"))
		self.assertEqual(str(subnet), "170.187.0.204/255.255.0.255")

	def test_subnet_full(self):
		subnet = IPAddressSubnet.from_bytes(bytes.fromhex("01 02 03 04 ff ff ff ff"))
		self.assertEqual(str(subnet), "1.2.3.4")

	def test_subnet_empty(self):
		subnet = IPAddressSubnet.from_bytes(bytes.fromhex("aa bb cc dd 00 00 00 00"))
		self.assertEqual(str(subnet), "170.187.204.221/0.0.0.0")

	def test_subnet_malformed(self):
		with self.assertRaises(InvalidIPAddressException):
			IPAddressSubnet(IPAddress(range(4)), IPAddress(range(16)))

	def test_subnet_parse(self):
		subnet = IPAddressSubnet.from_str("192.168.1.0/24")
		self.assertEqual(str(subnet.ip), "192.168.1.0")
		self.assertEqual(str(subnet.subnet), "255.255.255.0")

		subnet = IPAddressSubnet.from_str("192.168.1.0/255.255.255.0")
		self.assertEqual(str(subnet), "192.168.1.0/24")

	def test_cidr_create_ipv4(self):
		self.assertEqual(str(IPAddress.create_cidr_subnet(32)), "255.255.255.255")
		self.assertEqual(str(IPAddress.create_cidr_subnet(24)), "255.255.255.0")
		self.assertEqual(str(IPAddress.create_cidr_subnet(16)), "255.255.0.0")
		self.assertEqual(str(IPAddress.create_cidr_subnet(10)), "255.192.0.0")
		self.assertEqual(str(IPAddress.create_cidr_subnet(8)), "255.0.0.0")
		self.assertEqual(str(IPAddress.create_cidr_subnet(1)), "128.0.0.0")
		self.assertEqual(str(IPAddress.create_cidr_subnet(0)), "0.0.0.0")

	def test_cidr_contained(self):
		self.assertTrue(IPAddressSubnet.from_str("192.168.1.0/24").contains_ip(IPAddress.from_str("192.168.1.123")))
		self.assertTrue(IPAddressSubnet.from_str("192.168.1.0/24").contains_ip(IPAddress.from_str("192.168.1.0")))
		self.assertTrue(IPAddressSubnet.from_str("192.168.1.0/24").contains_ip(IPAddress.from_str("192.168.1.255")))
		self.assertFalse(IPAddressSubnet.from_str("192.168.1.0/24").contains_ip(IPAddress.from_str("192.168.2.255")))

	def test_subnet_overlap_simple(self):
		subnet = IPAddressSubnet.from_str("192.168.1.0/24")
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("0.0.0.0/0")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("192.168.1.34/32")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("192.168.1.0/24")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("192.168.0.0/16")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("192.0.0.0/8")))

		self.assertFalse(subnet.overlaps_subnet(IPAddressSubnet.from_str("192.168.2.34/32")))
		self.assertFalse(subnet.overlaps_subnet(IPAddressSubnet.from_str("192.5.1.0/24")))
		self.assertFalse(subnet.overlaps_subnet(IPAddressSubnet.from_str("193.168.0.0/16")))
		self.assertFalse(subnet.overlaps_subnet(IPAddressSubnet.from_str("255.0.0.0/8")))

		subnet = IPAddressSubnet.from_str("10.0.20.0/255.0.255.0")
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("0.0.0.0/0")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("10.3.20.4/32")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("10.0.20.0/32")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("10.255.20.255/32")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("10.0.0.0/8")))
		self.assertTrue(subnet.overlaps_subnet(IPAddressSubnet.from_str("0.0.20.0/0.0.255.0")))

	def test_subnet_ipv6_parse(self):
		subnet = IPAddressSubnet.from_str("abcd:1234:9999::/48")
		self.assertEqual(str(subnet.ip), "abcd:1234:9999::")
		self.assertEqual(str(subnet.subnet), "ffff:ffff:ffff::")



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import tempfile
from x509sak.tests import BaseTest
from x509sak.WorkDir import WorkDir
from x509sak.SubprocessExecutor import SubprocessExecutor

class CmdLineTestsCreateCA(BaseTest):
	def test_create_simple_ca(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"--BEGIN CERTIFICATE--", output)
			self.assertIn(b"--END CERTIFICATE--", output)
			self.assertIn(b"id-ecPublicKey", output)
			SubprocessExecutor([ "openssl", "ec", "-in", "root_ca/CA.key" ]).run()

	def test_create_simple_ca_2(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "-g", "ecc:secp256r1", "-s", "/CN=YepThatsTheCN", "-h", "sha512", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"--BEGIN CERTIFICATE--", output)
			self.assertIn(b"--END CERTIFICATE--", output)
			self.assertIn(b"id-ecPublicKey", output)
			self.assertIn(b"ecdsa-with-SHA512", output)
			self.assertIn(b"prime256v1", output)
			self.assertTrue((b"CN=YepThatsTheCN" in output) or (b"CN = YepThatsTheCN" in output))
			self.assertIn(b"X509v3 extensions", output)
			self.assertIn(b"CA:TRUE", output)
			self.assertIn(b"X509v3 Subject Key Identifier", output)
			SubprocessExecutor([ "openssl", "ec", "-in", "root_ca/CA.key" ]).run()

	def test_create_simple_ca_3(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "-g", "rsa:1024", "-s", "/CN=YepThats!TheCN", "-h", "sha1", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"--BEGIN CERTIFICATE--", output)
			self.assertIn(b"--END CERTIFICATE--", output)
			self.assertIn(b"rsaEncryption", output)
			self.assertIn(b"sha1WithRSAEncryption", output)
			self.assertIn(b"Public-Key: (1024 bit)", output)
			self.assertTrue((b"CN=YepThats!TheCN" in output) or (b"CN = YepThats!TheCN" in output))
			self.assertIn(b"X509v3 extensions", output)
			self.assertIn(b"CA:TRUE", output)
			self.assertIn(b"X509v3 Subject Key Identifier", output)
			SubprocessExecutor([ "openssl", "rsa", "-in", "root_ca/CA.key" ]).run()

	def test_create_nested_ca(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "this/is/a/root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "this/is/a/root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"--BEGIN CERTIFICATE--", output)
			self.assertIn(b"--END CERTIFICATE--", output)
			self.assertIn(b"id-ecPublicKey", output)
			SubprocessExecutor([ "openssl", "ec", "-in", "this/is/a/root_ca/CA.key" ]).run()

	def test_create_intermediate_ca(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "-s", "/CN=PARENT", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"--BEGIN CERTIFICATE--", output)
			self.assertIn(b"--END CERTIFICATE--", output)
			self.assertTrue((b"Issuer: CN=PARENT" in output) or (b"Issuer: CN = PARENT" in output))
			self.assertTrue((b"Subject: CN=PARENT" in output) or (b"Subject: CN = PARENT" in output))
			self.assertIn(b"id-ecPublicKey", output)
			SubprocessExecutor([ "openssl", "ec", "-in", "root_ca/CA.key" ]).run()

			self._run_x509sak([ "createca", "-p", "root_ca", "-s", "/CN=INTERMEDIATE", "intermediate_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "intermediate_ca/CA.crt" ]).run().stdout
			self.assertIn(b"--BEGIN CERTIFICATE--", output)
			self.assertIn(b"--END CERTIFICATE--", output)
			self.assertTrue((b"Issuer: CN=PARENT" in output) or (b"Issuer: CN = PARENT" in output))
			self.assertTrue((b"Subject: CN=INTERMEDIATE" in output) or (b"Subject: CN = INTERMEDIATE" in output))
			self.assertIn(b"id-ecPublicKey", output)
			SubprocessExecutor([ "openssl", "ec", "-in", "intermediate_ca/CA.key" ]).run()

	def test_subject_info(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "-s", "/CN=Elem00/OU=Elem01/C=DE/SN=Elem02/GN=Elem03/emailAddress=Elem04/title=Elem05/L=Elem06/stateOrProvinceName=Elem07/pseudonym=Elem08", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-subject", "-noout", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"DE", output)
			for eid in range(9):
				element = ("Elem%02d" % (eid)).encode("ascii")
				self.assertIn(element, output)

	def test_x509_extension(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "--extension", "nameConstraints=critical,permitted;DNS:foo.bar.com", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-noout", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"DNS:foo.bar.com", output)
			self.assertNotIn(b"pathlen", output)

			self._run_x509sak([ "createca", "--extension", "nameConstraints=critical,permitted;DNS:foo.bar.com", "--extension", "basicConstraints=critical,CA:TRUE,pathlen:123", "root_ca2" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-noout", "-in", "root_ca2/CA.crt" ]).run().stdout
			self.assertIn(b"DNS:foo.bar.com", output)
			self.assertIn(b"pathlen:123", output)

	def test_dont_overwrite_dir_except_force(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			os.makedirs("root_ca")
			with open("root_ca/foobar", "wb"):
				pass
			self._run_x509sak([ "createca", "root_ca" ], success_return_codes = [ 1 ])
			self.assertTrue(os.path.isfile("root_ca/foobar"))
			self._run_x509sak([ "createca", "--force", "root_ca" ])
			self.assertFalse(os.path.isfile("root_ca/foobar"))

	def test_serial_spec(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "-s", "/CN=PARENT", "--serial", "1234567", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"Serial Number: 1234567", output)

			self._run_x509sak([ "createca", "-s", "/CN=PARENT", "-f", "--serial", "0x1234567", "root_ca" ])
			output = SubprocessExecutor([ "openssl", "x509", "-text", "-in", "root_ca/CA.crt" ]).run().stdout
			self.assertIn(b"Serial Number: 19088743", output)

			# Catch error message
			output = self._run_x509sak([ "createca", "-p", "root_ca", "-s", "/CN=INTERMEDIATE", "--serial", "9876", "intermediate_ca" ], success_return_codes = [ 1 ]).stderr
			self.assertIn(b"specify certificate serial number", output)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyasn1.type import namedtype
from pyasn1.type import namedval
from pyasn1.type import tag
from pyasn1.type import univ, char
from pyasn1_modules import rfc2315, rfc3280

class ECPVer(univ.Integer):
	"""RFC 3279: Algorithms and Identifiers for the Internet X.509 Public Key Infrastructure Certificate and Certificate Revocation List (CRL) Profile

	ECPVer ::= INTEGER {ecpVer1(1)}
	"""
	namedValues = namedval.NamedValues(
		("ecpVer1", 1),
	)

class FieldID(univ.Sequence):
	"""RFC 3279: Algorithms and Identifiers for the Internet X.509 Public Key Infrastructure Certificate and Certificate Revocation List (CRL) Profile

	FieldID ::= SEQUENCE {
		fieldType   OBJECT IDENTIFIER,
		parameters  ANY DEFINED BY fieldType
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("fieldType", univ.ObjectIdentifier()),
		namedtype.NamedType("parameters", univ.Any()),
	)

class FieldElement(univ.OctetString):
	"""RFC 3279: Algorithms and Identifiers for the Internet X.509 Public Key Infrastructure Certificate and Certificate Revocation List (CRL) Profile

	FieldElement ::= OCTET STRING
	"""
	pass

class ECPoint(univ.OctetString):
	"""RFC 3279: Algorithms and Identifiers for the Internet X.509 Public Key Infrastructure Certificate and Certificate Revocation List (CRL) Profile

	ECPoint ::= OCTET STRING
	"""
	pass

class Curve(univ.Sequence):
	"""RFC 3279: Algorithms and Identifiers for the Internet X.509 Public Key Infrastructure Certificate and Certificate Revocation List (CRL) Profile

	Curve ::= SEQUENCE {
		a         FieldElement,
		b         FieldElement,
		seed      BIT STRING OPTIONAL
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("a", FieldElement()),
		namedtype.NamedType("b", FieldElement()),
		namedtype.OptionalNamedType("seed", univ.BitString()),
	)

class SpecifiedECDomain(univ.Sequence):
	"""RFC 3279: Algorithms and Identifiers for the Internet X.509 Public Key Infrastructure Certificate and Certificate Revocation List (CRL) Profile

	ECParameters ::= SEQUENCE {
		version   ECPVer,          -- version is always 1
		fieldID   FieldID,         -- identifies the finite field over which the curve is defined
		curve     Curve,           -- coefficients a and b of the elliptic curve
		base      ECPoint,         -- specifies the base point P on the elliptic curve
		order     INTEGER,         -- the order n of the base point
		cofactor  INTEGER OPTIONAL -- The integer h = #E(Fq)/n
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("version", ECPVer()),
		namedtype.NamedType("fieldID", FieldID()),
		namedtype.NamedType("curve", Curve()),
		namedtype.NamedType("base", ECPoint()),
		namedtype.NamedType("order", univ.Integer()),
		namedtype.OptionalNamedType("cofactor", univ.Integer()),
	)

class ECParameters(univ.Choice):
	"""RFC5480: Elliptic Curve Cryptography Subject Public Key Information

	ECParameters ::= CHOICE {
		namedCurve         OBJECT IDENTIFIER
		-- implicitCurve   NULL
		-- specifiedCurve  SpecifiedECDomain
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("namedCurve", univ.ObjectIdentifier()),
		namedtype.NamedType("implicitCurve", univ.Null()),
		namedtype.NamedType("specifiedCurve", SpecifiedECDomain()),
	)

class ECFieldParametersPrimeField(univ.Integer):
	"""X9.62-1998: Public Key Cryptography for the Financial Services Industry

	Prime-p ::= INTEGER        -- Finite field F(p), where p is an odd prime
	"""
	pass

class ECFieldParametersCharacteristicTwoField(univ.Sequence):
	"""X9.62-1998: Public Key Cryptography for the Financial Services Industry

	Characteristic-two ::= SEQUENCE {
		m INTEGER,			-- Field size 2^m
		basis CHARACTERISTIC-TWO.&id({BasisTypes}),
		parameters CHARACTERISTIC-TWO.&Type({BasisTypes}{@basis})
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("m", univ.Integer()),
		namedtype.NamedType("basis", univ.ObjectIdentifier()),
		namedtype.NamedType("parameters", univ.Any()),
	)

class ECFieldParametersCharacteristicTwoFieldTrinomial(univ.Integer):
	"""X9.62-1998: Public Key Cryptography for the Financial Services Industry

	x^m + x^k + 1

	Trinomial ::= INTEGER
	"""
	pass

class ECFieldParametersCharacteristicTwoFieldPentanomial(univ.Sequence):
	"""X9.62-1998: Public Key Cryptography for the Financial Services Industry

	x^m + x^k3 + x^k2 + x^k1 + 1

	Pentanomial ::= SEQUENCE {
		k1 INTEGER,
		k2 INTEGER,
		k3 INTEGER
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("k1", univ.Integer()),
		namedtype.NamedType("k2", univ.Integer()),
		namedtype.NamedType("k3", univ.Integer()),
	)

class ECPrivateKeyVersion(univ.Integer):
	namedValues = namedval.NamedValues(
		("ecPrivkeyVer1", 1),
	)

class ECPrivateKey(univ.Sequence):
	"""RFC5915: Elliptic Curve Private Key Structure

	ECPrivateKey ::= SEQUENCE {
		version        INTEGER { ecPrivkeyVer1(1) } (ecPrivkeyVer1),
		privateKey     OCTET STRING,
		parameters [0] ECParameters {{ NamedCurve }} OPTIONAL,
		publicKey  [1] BIT STRING OPTIONAL
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("version", ECPrivateKeyVersion()),
		namedtype.NamedType("privateKey", univ.OctetString()),
		namedtype.OptionalNamedType("parameters", ECParameters().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0))),
		namedtype.OptionalNamedType("publicKey", univ.BitString().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1))),
	)

class PFXVersion(univ.Integer):
	namedValues = namedval.NamedValues(
		("v3", 3),
	)

class MacData(univ.Sequence):
	"""RFC7292: PKCS #12: Personal Information Exchange Syntax v1.1

	MacData ::= SEQUENCE {
		mac        DigestInfo,
		macSalt    OCTET STRING,
		iterations INTEGER DEFAULT 1
		-- Note: The default is for historical reasons and its use is
		-- deprecated.
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("mac", rfc2315.DigestInfo()),
		namedtype.NamedType("macSalt", univ.OctetString()),
		namedtype.OptionalNamedType("iterations", univ.Integer()),
	)

class PFX(univ.Sequence):
	"""RFC7292: PKCS #12: Personal Information Exchange Syntax v1.1

	PFX ::= SEQUENCE {
		 version    INTEGER {v3(3)}(v3,...),
		 authSafe   ContentInfo,
		 macData    MacData OPTIONAL
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("version", PFXVersion()),
		namedtype.NamedType("authSafe", rfc2315.ContentInfo()),
		namedtype.OptionalNamedType("macData", MacData()),
	)

class DSASignature(univ.Sequence):
	"""RFC5912: New ASN.1 Modules for the Public Key Infrastructure Using X.509 (PKIX)

	DSA-Sig-Value ::= SEQUENCE {
		r  INTEGER,
		s  INTEGER
	}

	ECDSA-Sig-Value ::= SEQUENCE {
		r  INTEGER,
		s  INTEGER
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("r", univ.Integer()),
		namedtype.NamedType("s", univ.Integer()),
	)

class HashAlgorithm(rfc3280.AlgorithmIdentifier): pass
class MaskGenAlgorithm(rfc3280.AlgorithmIdentifier): pass
class TrailerField(univ.Integer): pass

class RSASSA_PSS_Params(univ.Sequence):
	"""RFC3447: Public-Key Cryptography Standards (PKCS) #1: RSA Cryptography Specifications Version 2.1

	RSASSA-PSS-params ::= SEQUENCE {
		hashAlgorithm      [0] HashAlgorithm    DEFAULT sha1,
		maskGenAlgorithm   [1] MaskGenAlgorithm DEFAULT mgf1SHA1,
		saltLength         [2] INTEGER          DEFAULT 20,
		trailerField       [3] TrailerField     DEFAULT trailerFieldBC
	}
	"""
	componentType = namedtype.NamedTypes(
		namedtype.DefaultedNamedType("hashAlgorithm", HashAlgorithm().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0))),
		namedtype.DefaultedNamedType("maskGenAlgorithm", MaskGenAlgorithm().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1))),
		namedtype.DefaultedNamedType("saltLength", univ.Integer().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 2))),
		namedtype.DefaultedNamedType("trailerField", TrailerField().subtype(explicitTag = tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 3))),
	)

class NetscapeCertificateType(univ.BitString):
	namedValues = namedval.NamedValues(
		("client", 0),
		("server", 1),
		("email", 2),
		("objsign", 3),
		("reserved", 4),
		("sslCA", 5),
		("emailCA", 6),
		("objCA", 7),
	)

class RelaxedDisplayText(univ.Choice):
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("ia5String", char.IA5String()),
		namedtype.NamedType("visibleString", char.VisibleString()),
		namedtype.NamedType("bmpString", char.BMPString()),
		namedtype.NamedType("utf8String", char.UTF8String()),
	)

class RelaxedNoticeReference(univ.Sequence):
	componentType = namedtype.NamedTypes(
		namedtype.NamedType("organization", RelaxedDisplayText()),
	    namedtype.NamedType("noticeNumbers", univ.SequenceOf(componentType = univ.Integer())),
	)

class RelaxedUserNotice(univ.Sequence):
	componentType = namedtype.NamedTypes(
		namedtype.OptionalNamedType("noticeRef", RelaxedNoticeReference()),
		namedtype.OptionalNamedType("explicitText", RelaxedDisplayText()),
	)

class RelaxedCPSuri(RelaxedDisplayText): pass



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.Tools import JSONTools
from x509sak.OID import OID
from x509sak.Exceptions import LazyDeveloperException, CurveNotFoundException, InvalidInputException
from x509sak.ECCMath import EllipticCurve

class CurveDB():
	_DB_DATA = None
	_OID_BY_NAME = None

	def __init__(self):
		if self._DB_DATA is None:
			self._DB_DATA = self._load_db_data()
		if self._OID_BY_NAME is None:
			self._OID_BY_NAME = { curve["name"]: oid for (oid, curve) in self._DB_DATA.items() }

	@staticmethod
	def _load_db_data():
		json_data = JSONTools.load_internal("x509sak.data.ecc", "curves.json")
		result = { OID.from_str(oid): parameters for (oid, parameters) in json_data.items() }
		for (oid, parameters) in result.items():
			parameters["oid"] = oid
		return result

	def lookup(self, oid = None, name = None, on_error = "none"):
		assert(on_error in [ "none", "raise" ])

		# Check that either OID or name must be given, not both
		if (oid is None) and (name is None):
			raise InvalidInputException("Lookup from curve database needs either OID or name of curve to look up.")
		if (oid is not None) and (name is not None):
			raise InvalidInputException("Lookup from curve database needs either OID or name of curve to look up, not both. Given: OID %s and name %s." % (oid, name))

		if oid is not None:
			curve_data = self._DB_DATA.get(oid)
		else:
			curve_data = self._DB_DATA.get(self._OID_BY_NAME.get(name))
		if (curve_data is None) and (on_error == "raise"):
			if oid is not None:
				raise CurveNotFoundException("No such curve with OID %s in database." % (oid))
			else:
				raise CurveNotFoundException("No such curve with name %s in database." % (name))
		return curve_data

	def instantiate(self, oid = None, name = None):
		curve_data = self.lookup(oid = oid, name = name, on_error = "raise")

		domain = curve_data["domain"]
		metadata = {
			"name":		curve_data["name"],
			"oid":		curve_data["oid"],
		}
		handling_class = EllipticCurve.get_class_for_curvetype(curve_data["curvetype"])
		if handling_class is None:
			raise LazyDeveloperException(NotImplemented, curve_data["curvetype"])
		return handling_class(metadata = metadata, **domain)

	def lookup_by_params(self, curve):
		for curvedata in self._DB_DATA.values():
			if curvedata["curvetype"] == curve.curvetype:
				instance = self.instantiate(name = curvedata["name"])
				if instance == curve:
					return instance
		return None

	def __iter__(self):
		return iter(self._DB_DATA)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import math
from x509sak.AlgorithmDB import Cryptosystems
from x509sak.NumberTheory import NumberTheory
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, LiteratureReference
from x509sak.ECCMath import PrimeFieldEllipticCurve, BinaryFieldEllipticCurve
from x509sak.CurveDB import CurveDB

@BaseEstimator.register
class ECCSecurityEstimator(BaseEstimator):
	_ALG_NAME = "ecc"

	def _judge_curve_cofactor(self, curve):
		judgements = SecurityJudgements()

		if curve.h is None:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Cofactor_Missing, "Curve cofactor h is not present in explicit domain parameter encoding. This is allowed, but highly unusual.", commonness = Commonness.HIGHLY_UNUSUAL)
		else:
			if curve.h <= 0:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_Cofactor_Invalid, "Curve cofactor h = %d is zero or negative. This is invalid." % (curve.h), bits = 0, commonness = Commonness.HIGHLY_UNUSUAL)
			elif curve.h > 8:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_Cofactor_Large, "Curve cofactor is unusually large, h = %d. This is an indication the curve has non-ideal cryptographic properties; would expect h <= 8." % (curve.h), commonness = Commonness.HIGHLY_UNUSUAL)

			if curve.curvetype == "prime":
				field_size = curve.p
			elif curve.curvetype == "binary":
				# TODO: For GF(p), the number of group elements is simply p.
				# For GF(2^m), I'm fairly certiain it is the reduction
				# polynomial (essentially, it's modular polynomial arithmetic).
				# Not 100% sure though. Verify.
				field_size = curve.int_poly

			# Hasse Theorem on Elliptic Curves:
			# p - (2 * sqrt(p)) + 1 <= #E(F_p) <= p + (2 * sqrt(p)) + 1
			# #E(F_p) = n h
			# h >= (p - (2 * sqrt(p)) + 1) / n
			# h <= (p + (2 * sqrt(p)) + 1) / n
			sqrt_p = NumberTheory.isqrt(field_size)
			hasse_h_min = (field_size - (2 * (sqrt_p + 1)) + 1) // curve.n
			hasse_h_max = (field_size + (2 * (sqrt_p + 1)) + 1) // curve.n

			if not (hasse_h_min <= curve.h <= hasse_h_max):
				literature = LiteratureReference(author = "Helmut Hasse", title = "Zur Theorie der abstrakten elliptischen Funktionenkörper. I, II & III", year = 1936, source = "Crelle's Journal")
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_Cofactor_OutsideHasseBound, "Curve cofactor h = %d is outside the Hasse bound (%d <= h <= %d). Cofactor therefore is invalid." % (curve.h, hasse_h_min, hasse_h_max), bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, literature = literature)
		return judgements

	def _judge_curve_embedding_degree(self, curve):
		d = 1
		for k in range(1, 50 + 1):
			d = (d * curve.p) % curve.n
			if d == 1:
				if (k == 1) or (NumberTheory.is_probable_prime(curve.n)):
					fail_text = "k = %d" % (k)
				else:
					fail_text = "k <= %d" % (k)
				literature = LiteratureReference(author = [ "Alfred Menezes", "Scott Vanstone", "Tatsuaki Okamoto" ], title = "Reducing Elliptic Curve Logarithms to Logarithms in a Finite Field", year = 1991, source = "ACM")
				return SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_LowEmbeddingDegree, "This curve has low embedding degree (%s), it fails the MOV condition. It can be compromised using the probabilistic polynomial-time MOV attack." % (fail_text), bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, literature = literature)
		return None

	def _judge_prime_field_curve(self, curve):
		judgements = SecurityJudgements()

		judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_PrimeField, "Prime finite field elliptic curve is used.", commonness = Commonness.COMMON)

		if curve.h is not None:
			# TODO: We might be able to guess the cofactor because of the Hasse bound.
			# E(Fp) = p + 1 - t
			# t = p - E(Fp) + 1
			EFp = curve.n * curve.h
			trace = curve.p - EFp + 1
			if trace == 0:
				literature = LiteratureReference(author = [ "Alfred Menezes", "Scott Vanstone", "Tatsuaki Okamoto" ], title = "Reducing Elliptic Curve Logarithms to Logarithms in a Finite Field", year = 1991, source = "ACM")
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_SupersingularCurve, "This curve is supersingular, trace is zero. The curve can be attacked using the probabilistic polynomial-time MOV attack.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, literature = literature)
			elif trace == 1:
				literature = LiteratureReference(author = [ "Nigel P. Smart" ], title = "The discrete logarithm problem on elliptic curves of trace one", year = 1997, month = 10, source = "HP Laboratories Bristol")
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_AnomalousCurve, "This curve is anomalous, #E(F_p) is equal to p. The curve can be attacked in linear time.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, literature = literature)

		p = ((4 * (curve.a ** 3)) + (27 * (curve.b ** 2))) % curve.p
		if p == 0:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_SingularCurve, "This curve is singular, 4a³ + 27b² = 0 mod p. Mathematical assumptions about the structure of the curve do not hold.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL)

		return judgements

	def _judge_binary_field_curve(self, curve):
		judgements = SecurityJudgements()

		literature = LiteratureReference(author = [ "Steven D. Galbraith", "Shishay W. Gebregiyorgis" ], title = "Summation polynomial algorithms for elliptic curves in characteristic two", year = 2014, source = "Progress in Cryptology -- INDOCRYPT 2014; LNCS 8885")
		judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_BinaryField, "Binary finite field elliptic curve is used. Recent advances in cryptography show there might be efficient attacks on such curves, hence it is recommended to use prime-field curves instead.", commonness = Commonness.UNUSUAL, literature = literature)

		EFpm = curve.n * curve.h
		if (EFpm % 2) == 1:
			# Supersingular: #E(F_p^m) = 1 mod p
			literature = LiteratureReference(author = [ "Alfred Menezes", "Scott Vanstone", "Tatsuaki Okamoto" ], title = "Reducing Elliptic Curve Logarithms to Logarithms in a Finite Field", year = 1991, source = "ACM")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_SupersingularCurve, "This curve is supersingular, #E(F_p^m) = 1 mod p. The curve can be attacked using an probabilistic polynomial-time MOV attack.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, literature = literature)

		if not NumberTheory.is_probable_prime(curve.m):
			literature = LiteratureReference(author = [ "Jeffrey Hoffstein", "Jill Pipher", "Joseph Silverman" ], title = "An Introduction to Mathematical Cryptography", year = 2008, source = "Springer")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_WeilDescent, "Binary finite field elliptic curve has a field size that is non-primem, F(2^%d). Weil Descent attacks could be successful.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, literature = literature)

		return judgements

	def _check_explicit_curve_params(self, curve):
		judgements = SecurityJudgements()

		curve_db = CurveDB()
		known_curve = curve_db.lookup_by_params(curve)
		if known_curve is None:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Name_UnknownExplicit, "Explicit curve domain parameter encoding with domain parameters that are not present in the database. Highly suspect, convervatively rating as broken security.", commonness = Commonness.HIGHLY_UNUSUAL, bits = 0)
		else:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Name_UnusedName, "Explicit curve domain parameter encoding is used; curve domain parameters are equal to curve %s (OID %s). Recommend switching to that named curve." % (known_curve.name, known_curve.oid))

		judgements += self._judge_curve_cofactor(curve)

		if curve.curvetype == "binary":
			if len(curve.poly) != len(set(curve.poly)):
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_BinaryField_DuplicatePolynomialPower, "ECC field polynomial contains duplicate powers: %s -- Conservatively rating as broken security." % (str(curve.poly)), commonness = Commonness.HIGHLY_UNUSUAL, bits = 0)

			for custom_coeff in curve.poly[1 : -1]:
				if custom_coeff <= 1:
					judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_BinaryField_InvalidPolynomialPower, "ECC field polynomial contains x^%d where it would be expected to see a power of two or higher." % (custom_coeff), commonness = Commonness.HIGHLY_UNUSUAL, bits = 0)
				elif custom_coeff >= curve.m:
					judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_BinaryField_InvalidPolynomialPower, "ECC field polynomial contains x^%d where it would be expected to see a power of less than x^m (i.e., x^%d)." % (custom_coeff, curve.m), commonness = Commonness.HIGHLY_UNUSUAL, bits = 0)
		elif curve.curvetype == "prime":
			# Only check embedding degree for non-named (explicit) curves; it's
			# a comparatively expensive test and we assume that curves in the
			# database are all cryptographically sound.
			judgements += self._judge_curve_embedding_degree(curve)

			# This isn't computationally expensive, but we also assume database
			# curves are cryptographically sound.
			judgements += self._judge_curve_cofactor(curve)

		return judgements

	def _check_explicit_curve_encoding(self, pubkey):
		judgements = SecurityJudgements()
		param_decoding = pubkey.key.decoding_details[0]
		if param_decoding.asn1 is not None:
			seed_bitstring = param_decoding.asn1.getComponent()["curve"]["seed"]
			if seed_bitstring.hasValue():
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Seed_Present, "Explicitly encoded public key contains seed. This is unnecessary, but not uncommon.")

				if len(seed_bitstring) == 0:
					judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Seed_Empty, "Explicitly encoded public key contains seed, but that seed is empty.", commonness = Commonness.UNUSUAL)
				elif len(seed_bitstring) > 256:
					judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Seed_Long, "Explicitly encoded public key contains long seed (%d bits)." % (len(seed_bitstring)), commonness = Commonness.UNUSUAL)

				if (len(seed_bitstring) % 8) != 0:
					judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Seed_NoByteString, "Explicitly encoded public key seed has uneven bit length (%d bits)." % (len(seed_bitstring)), commonness = Commonness.HIGHLY_UNUSUAL)

		return judgements

	def _analyze_curve(self, pubkey):
		judgements = SecurityJudgements()
		curve = pubkey.curve

		# Check that the encoded public key point is on curve first
		Q = curve.point(pubkey.x, pubkey.y)
		if not Q.on_curve():
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_PublicKeyPoint_NotOnCurve, "Public key point Q is not on the underlying curve %s." % (pubkey.curve), bits = 0)

		# Check that the encoded public key is not Gx
		if Q.x == curve.Gx:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_PublicKeyPoint_IsGenerator, "Public key point Q_x is equal to generator G_x on curve %s." % (pubkey.curve), bits = 0)

		# We assume, completely out-of-the-blue and worst-case estimate, 32
		# automorphisms that could be present for any curve (see Duursma et
		# al., "Speeding up the discrete log computation on curves with
		# automorphisms"). Therefore, for a given order n, we estimate the
		# complexity in bits as:
		#
		# b = log2(sqrt(n / 32)) = (log2(n) / 2) - 2.5
		approx_curve_order_bits = curve.order_bits
		bits_security = (approx_curve_order_bits / 2) - 2.5

		# We then take into account anomalous binary curves (Koblitz curves) as
		# well and use the approximations of Wiener/Zuccherato ("Faster Attacks
		# on Elliptic Curve Cryptosystems")
		literature = LiteratureReference(author = [ "Michael J. Wiener", "Robert J. Zuccherato" ], title = "Faster Attacks on Elliptic Curve Cryptosystems", year = 1999, source = "Selected Areas in Cryptography 1998; LNCS 1556")
		if isinstance(curve, BinaryFieldEllipticCurve) and curve.is_koblitz:
			speedup = math.sqrt(2 * curve.m)
			bits_security -= math.log(speedup, 2)
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_KoblitzCurve, "Binary field Koblitz curves (anomalous binary curves) have more efficient attacks than their non-anomalous binary curves; in this case improving attack performance by a factor of ~%.1f." % (speedup), commonness = Commonness.UNUSUAL, literature = literature)

		if isinstance(curve, PrimeFieldEllipticCurve) and curve.is_koblitz:
			# The math here is a bit shady. Firstly, Koblitz curves over F_p
			# only mean there's an efficiently computable endomorphism (e.g.,
			# R. Gallant (1999); "Faster elliptic curve cryptography using
			# efficient endomorphisms"). We do not check for that, however, but
			# instead rely on dull "b = 0 and a is small" check.
			# Additionally, Wiener and Zuccherato describe curves of form
			# y^2 = x^3 - ax or y^2 = x^3 + b (which, for our a/b check, is not
			# the case) and, for the latter, describe a sqrt(6) speedup. We
			# just take that as is, knowing full well it's just guesswork.
			speedup = math.sqrt(6)
			bits_security -= math.log(speedup, 2)
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_CurveProperty_KoblitzCurve, "Prime field Koblitz curves might have more efficient attacks than non-Koblitz curves. In this case, attack performance improves roughly by a factor of ~%.1f." % (speedup), commonness = Commonness.UNUSUAL, literature = literature)

		bits_security = math.floor(bits_security)
		judgements += self.algorithm("bits").analyze(JudgementCode.X509Cert_PublicKey_ECC_CurveOrderInBits, bits_security)

		# Check if the affine X/Y coordinates of the public key are about the
		# same length as the curve order. If randomly generated, both X and Y
		# should be about the same bitlength as the generator order and the
		# hamming weight should be roughly half of the bitlength of the curve
		# order.
		hweight_analysis = NumberTheory.hamming_weight_analysis(pubkey.x, min_bit_length = curve.field_bits)
		if not hweight_analysis.plausibly_random:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_PublicKeyPoint_X_BitBiasPresent, "Hamming weight of public key field element's X coordinate is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		hweight_analysis = NumberTheory.hamming_weight_analysis(pubkey.y, min_bit_length = curve.field_bits)
		if not hweight_analysis.plausibly_random:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_PublicKeyPoint_Y_BitBiasPresent, "Hamming weight of public key field element's Y coordinate is %d at bitlength %d, but expected a weight between %d and %d when randomly chosen; this is likely not coincidential." % (hweight_analysis.hweight, hweight_analysis.bitlen, hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		if isinstance(curve, BinaryFieldEllipticCurve):
			judgements += self._judge_binary_field_curve(curve)
		elif isinstance(curve, PrimeFieldEllipticCurve):
			judgements += self._judge_prime_field_curve(curve)

		return judgements

	def analyze(self, pubkey):
		judgements = SecurityJudgements()

		if pubkey.curve_source != "implicitCurve":
			# Can only check most of the things when there's a curve we can actually check
			judgements += self._analyze_curve(pubkey)

		if pubkey.curve_source == "namedCurve":
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Name_NamedCurve, "Curve referenced by name in parameter field.", commonness = Commonness.COMMON)
		elif pubkey.curve_source == "specifiedCurve":
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Name_ExplicitCurve, "Curve uses explicit encoding for domain parameters. Typically, named curves are used; explicit encoding of domain parameters is not recommended and may be rejected by implementations for simplicity reasons.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)
			judgements += self._check_explicit_curve_encoding(pubkey)
			judgements += self._check_explicit_curve_params(pubkey.curve)
		else:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_ECC_DomainParameters_Name_ImplicitCurve, "Curve uses implicit curve for domain parameters. This is highly uncommon practice and strongly discouraged.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.LIMITED_SUPPORT)

		result = {
			"specific":	{ },
			"security":			judgements,
		}

		if pubkey.pk_alg.value.cryptosystem == Cryptosystems.ECC_ECDSA:
			result["cryptosystem"] = "ecc/ecdsa"
		elif pubkey.pk_alg.value.cryptosystem == Cryptosystems.ECC_EdDSA:
			result["cryptosystem"] = "ecc/eddsa"
		else:
			raise NotImplementedError("ECC estimator currently not fit to analyze a %s pubkey." % (pubkey.pk_alg.value.cryptosystem.name))

		return result
